# -*- coding: utf-8 -*-
from enum import Enum, unique
from functools import partial, wraps
from itertools import count, product, repeat, takewhile
from math import cos, isclose, radians, sin, sqrt
from typing import Any, Generator, Iterable, Iterator, Tuple, Union

import numpy as np
from numpy.linalg import norm
from numpy.typing import ArrayLike

from .affine import change_basis_mesh, change_of_basis


def primed(gen: Generator) -> Generator:
    """
    Decorator that primes a generator function, i.e. runs the function
    until the first ``yield`` statement. Useful in cases where there
    are preliminary checks when creating the generator.
    """

    @wraps(gen)
    def primed_gen(*args, **kwargs):
        generator = gen(*args, **kwargs)
        next(generator)
        return generator

    return primed_gen


def matmulrow(matrix: np.ndarray, arr: np.ndarray) -> np.ndarray:
    """Row-wise matrix multiplication."""
    return np.transpose(matrix @ arr.T)


@unique
class LatticeSystem(Enum):
    """
    Lattice system enumeration.

    Equivalent to Crystal family
    except that the hexagonal crystal family is split between
    the rhombohedral system and hexagonal system.
    """

    triclinic = 1
    monoclinic = 2
    orthorhombic = 3
    tetragonal = 4
    rhombohedral = 5
    hexagonal = 6
    cubic = 7


class Lattice:
    """
    Container class for lattice information and manipulations.

    Instances can also be create from the standard 'three lengths and angles'
    parameters via ``Lattice.from_parameters``:

    Parameters
    ----------
    lattice_vectors: iterable of `~numpy.ndarray`, shape (3,)
        Lattice vectors.
    """

    def __init__(self, lattice_vectors: ArrayLike, **kwargs):
        a1, a2, a3 = lattice_vectors
        self.a1 = np.asarray(a1, dtype=float)
        self.a2 = np.asarray(a2, dtype=float)
        self.a3 = np.asarray(a3, dtype=float)
        super().__init__(**kwargs)

    def __repr__(self) -> str:
        a, b, c, alpha, beta, gamma = self.lattice_parameters
        return f"< Lattice object with parameters {a:.3f}Å, {b:.3f}Å, {c:.3f}Å, {alpha:.2f}°, {beta:.2f}°, {gamma:.2f}° >"

    def __hash__(self) -> int:
        return hash(self.lattice_parameters)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Lattice):
            return np.allclose(self.lattice_vectors, other.lattice_vectors, atol=1e-3)
        return NotImplemented

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """Returns a 3x3 float array in which each row is a lattice vector"""
        return np.array(self.lattice_vectors, *args, **kwargs)

    @classmethod
    def from_parameters(
        cls, a: float, b: float, c: float, alpha: float, beta: float, gamma: float
    ):
        """
        Create a Lattice instance from three lengths and angles.

        Parameters
        ----------
        a, b, c : floats
            Lattice vectors lengths [Å]
        alpha, beta, gamma : floats
            Angles between lattice vectors [deg]

        Raises
        ------
        ValueError : if lattice parameters are invalid.
        """
        return cls(lattice_vectors_from_parameters(a, b, c, alpha, beta, gamma))

    @property
    def lattice_parameters(self) -> Tuple[float, float, float, float, float, float]:
        """Lattice parameters as three lengths [Å] and three angles [degrees]."""
        a, b, c = norm(self.a1), norm(self.a2), norm(self.a3)
        alpha = np.arccos(np.vdot(self.a2, self.a3) / (b * c))
        beta = np.arccos(np.vdot(self.a1, self.a3) / (a * c))
        gamma = np.arccos(np.vdot(self.a1, self.a2) / (a * b))
        return a, b, c, np.rad2deg(alpha), np.rad2deg(beta), np.rad2deg(gamma)

    @property
    def lattice_system(self) -> "LatticeSystem":
        """One of the seven lattice system, returned in the form of the :class:`LatticeSystem` enumeration."""
        return lattice_system(*self.lattice_parameters, atol=5e-2)

    @property
    def volume(self) -> float:
        """Lattice cell volume Angtroms cubed"""
        return np.dot(self.a1, np.cross(self.a2, self.a3))

    @property
    def lattice_vectors(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Iterable of lattice vectors"""
        return self.a1, self.a2, self.a3

    @lattice_vectors.setter
    def lattice_vectors(self, vectors: ArrayLike):
        self.a1, self.a2, self.a3 = vectors

    @property
    def reciprocal(self) -> "Lattice":
        """Reciprocal lattice"""
        return Lattice(lattice_vectors=self.reciprocal_vectors)

    @property
    def reciprocal_vectors(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Reciprocal lattice vectors, defined as:

        .. math::

            b_i = 2 \\pi \\frac{a_j \\times a_k}{v}

        For :math:`v` the unit cell volume.
        """
        cell_volume = self.volume
        b1 = 2 * np.pi * np.cross(self.a2, self.a3) / cell_volume
        b2 = 2 * np.pi * np.cross(self.a3, self.a1) / cell_volume
        b3 = 2 * np.pi * np.cross(self.a1, self.a2) / cell_volume
        return b1, b2, b3

    @property
    def periodicity(self) -> Tuple[float, float, float]:
        """
        Crystal periodicity in x, y and z direction from the lattice constants.
        This is effectively a bounding cube for the unit cell, which is itself a unit cell.

        Returns
        -------
        x, y, z : float
            Periodicity in Angstroms along the x, y, and z direction, respectively.
        """
        # Add the absolute value of the component of every lattice vector
        # along the three euclidian vectors, which is effectively the sum of
        # absolutes of columns
        lv = np.abs(np.array(self.lattice_vectors))
        return tuple(lv.sum(axis=0))

    def scattering_vector(self, reflection: ArrayLike) -> np.ndarray:
        """
        Scattering vector from Miller indices.

        .. versionchanged:: 1.3
           Can now operate on tables of reflections, where every
           reflection is a row.

        Parameters
        ----------
        reflection : array_like, shape (3,) or (N, 3)
            Miller indices.

        Returns
        -------
        G : ndarray, shape (3,) or (N, 3)
            Scattering vector in :math:`A^{-1}`.
        """
        COB = change_of_basis(basis1=self.reciprocal_vectors, basis2=np.eye(3))
        return matmulrow(COB, np.asarray(reflection))

    def miller_indices(self, scattering_vector: ArrayLike) -> np.ndarray:
        """
        Miller indices from scattering vector components.

        .. versionchanged:: 1.3
           Can now operate on tables of vectors, where every
           vector is a row.

        Parameters
        ----------
        scattering_vector : array_like, shape (3,) or (N, 3)
            Scattering vector in :math:`A^{-1}`.

        Returns
        -------
        reflection : `~numpy.ndarray`, shape (3,) or (N, 3)
            Miller indices.
        """
        COB = change_of_basis(basis1=np.eye(3), basis2=self.reciprocal_vectors)
        return matmulrow(COB, np.asarray(scattering_vector))

    @staticmethod
    def frac_mesh(
        *xi,
        indexing: str = "xy",  # Not using Literal type because it landed in Python 3.8
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Coordinate arrays for fractional coordinates.

        Parameters
        ----------
        x1, x2, x3 : `~numpy.ndarray`, shape (N,)
            1d coordinate vectors. If only ``x1`` is provided, it is assumed
            that ``x1 = x2 = x3``. Otherwise, three coordinate vectors are expected.
        indexing : str, {'ij', 'xy'}
            Cartesian (‘xy’, default) or matrix (‘ij’) indexing of output.

        Returns
        -------
        out1, out2, out3 : `~numpy.ndarray`
            Fractional coordinate arrays.

        Raises
        ------
        ValueError : if number of input vectors is neither 1 nor 3.

        See Also
        --------
        numpy.meshgrid : Coordinate arrays from coordinate vectors
        Lattice.mesh : Real-space coordinate arrays from fractional coordinate vectors
        """
        if len(xi) == 1:
            xi = tuple(repeat(xi[0], times=3))
        elif len(xi) != 3:
            raise ValueError(
                f"1 or 3 coordinate arrays are required, but received {len(xi)}"
            )

        return np.meshgrid(*xi, indexing=indexing)

    def mesh(
        self,
        *xi: np.ndarray,
        indexing: str = "xy",  # Not using Literal type because it landed in Python 3.8
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Cartesian coordinate arrays from fractional coordinate vectors.

        Parameters
        ----------
        x1, x2, x3 : `~numpy.ndarray`, shape (N,)
            1d coordinate vectors in fractional coordinates.
            If only ``x1`` is provided, it is assumed that ``x1 = x2 = x3``.
            Otherwise, three coordinate vectors are expected.
        indexing : str, {'ij', 'xy'}
            Cartesian (‘xy’, default) or matrix (‘ij’) indexing of output.

        Returns
        -------
        out1, out2, out3 : `~numpy.ndarray`
            Real-space (cartesian) coordinate arrays.

        Raises
        ------
        ValueError : if number of input vectors is neither 1 nor 3.

        See Also
        --------
        numpy.meshgrid : Coordinate arrays from coordinate vectors
        Lattice.frac_mesh : Coordinate arrays for fractional coordinates
        """
        return change_basis_mesh(
            *self.frac_mesh(*xi, indexing=indexing),
            basis1=np.array(self.lattice_vectors),
            basis2=np.eye(3),
        )

    # Primed generators allows for checks on creation
    # All lines of code before the first 'yield' will run at first
    @primed
    def bounded_reflections(
        self, bound: float, min_bound: float = 0
    ) -> Iterator[Tuple[int, int, int]]:
        """
        Generates reflections (hkl) with norm(G) <= bound

        Parameters
        ----------
        bound : float
            Maximal scattering vector norm [:math:`A^{-1}`].
        min_bound : float, optional
            Minimal scattering vector norm [:math:`A^{-1}`].

            .. versionadded:: 1.2.0

        Yields
        ------
        reflection : 3-tuple of ints
            Miller indices of a bounded reflection.

        Raises
        ------
        ValueError : if `bound` is smaller than `min_bound`.

        Examples
        --------
        >>> cryst = Crystal.from_database('C')
        >>> refls = cryst.bounded_reflections(1.5) # 1.5 inverse Angstroms
        >>> list(refls)
        [(0, 0, -1), (0, 0, 0), (0, 0, 1)]
        """
        if bound < min_bound:
            raise ValueError(f"Bound {bound} is smaller than the minimum {min_bound}.")

        # Determine the maximum index such that (i00) family is still within data limits
        # This provides a (large) upper bound so that we are sure that the overall filtering will terminate
        bounded = lambda i: any(
            [np.linalg.norm(i * b) <= bound for b in self.reciprocal_vectors]
        )

        max_index = max(takewhile(bounded, count(0)))
        extent = range(-max_index, max_index + 1)
        refls = product(extent, repeat=3)

        yield  # Priming the generator ends here

        # Generalized hypotenuse
        def _hypot(*args):
            return sqrt(sum(map(lambda i: i**2, args)))

        # The above bound was only a first pass. We can refine further
        in_bounds = (
            lambda refl: min_bound <= _hypot(*self.scattering_vector(refl)) <= bound
        )
        yield from filter(in_bounds, refls)


def lattice_vectors_from_parameters(
    a: float, b: float, c: float, alpha: float, beta: float, gamma: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """

    Parameters
    ----------
    a, b, c : floats
        Lattice vectors lengths [Å]
    alpha, beta, gamma : floats
        Angles between lattice vectors [deg]
    """
    # The algorithm below is a refactoring of the
    # one made available in Diffpy.Structure
    alpha, beta, gamma = map(radians, (alpha, beta, gamma))

    # The unit cell volume when a = b = c = 1.
    unit_volume = sqrt(
        1.0
        + 2.0 * cos(alpha) * cos(beta) * cos(gamma)
        - cos(alpha) ** 2
        - cos(beta) ** 2
        - cos(gamma) ** 2
    )

    # reciprocal lattice
    a_recip = sin(alpha) / (a * unit_volume)
    cos_gamma_recip = (cos(alpha) * cos(beta) - cos(gamma)) / (sin(alpha) * sin(beta))
    sin_gamma_recip = sqrt(1 - cos_gamma_recip**2)

    a1 = np.asarray(
        [1 / a_recip, -cos_gamma_recip / sin_gamma_recip / a_recip, cos(beta) * a], dtype=float
    )
    a2 = np.asarray([0, b * sin(alpha), b * cos(alpha)], dtype=float)
    a3 = np.asarray([0, 0, c], dtype=float)
    return a1, a2, a3


# TODO: also determine body-centered, primitive, face-centered, etc.
#       https://en.wikipedia.org/wiki/Bravais_lattice#Bravais_lattices_in_3_dimensions
def lattice_system(
    a: float,
    b: float,
    c: float,
    alpha: float,
    beta: float,
    gamma: float,
    atol: float = 1e-2,
) -> LatticeSystem:
    """
    Determine the lattice system. All cyclic permutations are checked,
    so that no convention on ordering of lattice parameters is assumed.

    Parameters
    ----------
    a, b, c : floats
        Lattice vectors lengths [Å]
    alpha, beta, gamma : floats
        Angles between lattice vectors [deg]
    atol : float, optional
        Absolute tolerance (in Angstroms)

    Returns
    -------
    system : LatticeSystem
        One of the seven lattice system.
    """
    lengths, angles = (a, b, c), (alpha, beta, gamma)

    angleclose = partial(isclose, abs_tol=1)
    lengthclose = partial(isclose, abs_tol=atol)

    lengths_equal = all(lengthclose(length, a) for length in lengths)
    angles_equal = all(angleclose(angle, alpha) for angle in angles)

    # Checking for monoclinic lattice system is generalized
    # to the case where (a, b, c) can be cycled
    # i.e. a != c and beta != 90
    #   or b != c and alpha != 90
    #   or a != b and gamma != 90
    for clengths, cangles in zip(cyclic(lengths), cyclic(angles)):
        (l1, _, l3), (a1, a2, a3) = clengths, cangles
        if (
            (not lengthclose(l1, l3))
            and angleclose(a1, 90)
            and angleclose(a3, 90)
            and (not angleclose(a2, 90))
        ):
            return LatticeSystem["monoclinic"]

    if lengths_equal and angles_equal:
        if angleclose(alpha, 90):
            return LatticeSystem["cubic"]
        else:
            return LatticeSystem["rhombohedral"]

    # Special note : technically, a hexagonal lattice system
    # could have all three lengths equal
    elif lengths_equal and (not angles_equal):
        if any(isclose(angle, 120) for angle in angles) and (
            sum(isclose(i, 90) for i in angles) == 2
        ):
            return LatticeSystem["hexagonal"]

    # At this point, two lengths are equal at most
    elif _two_equal(lengths, atol=atol):
        if angles_equal and angleclose(alpha, 90):
            return LatticeSystem["tetragonal"]

        elif any(isclose(angle, 120) for angle in angles) and (
            sum(isclose(i, 90) for i in angles) == 2
        ):
            return LatticeSystem["hexagonal"]

    # At this point, all lengths are unequal
    elif angles_equal and angleclose(alpha, 90):
        return LatticeSystem["orthorhombic"]

    else:
        return LatticeSystem["triclinic"]


def _two_equal(iterable, atol):
    """Returns True if and only if two items from an iterable are equal"""
    iterable = tuple(iterable)
    for i in iterable:
        if sum(isclose(i, l, abs_tol=atol) for l in iterable) == 2:
            return True
    return False


def cyclic(iterable: Iterable[Any]) -> Iterable[Iterable[Any]]:
    """
    Yields cyclic permutations of an iterable.

    Examples
    --------
    >>> list(cyclic((1,2,3)))
    [(1,2,3), (2,3,1), (3,1,2)]
    """
    iterable = tuple(iterable)
    n = len(iterable)
    yield from (tuple(iterable[i - j] for i in range(n)) for j in range(n))



# -*- coding: utf-8 -*-

from collections import OrderedDict, namedtuple
from copy import deepcopy
from enum import Enum, auto, unique
from multiprocessing.dummy import Array
from typing import Any, Callable, Dict, Optional, Union

import numpy as np
from numpy.typing import ArrayLike

from .affine import change_of_basis, transform
from .atom_data import (
    ELEM_TO_MAGMOM,
    ELEM_TO_MASS,
    ELEM_TO_NAME,
    ELEM_TO_NUM,
    NAME_TO_ELEM,
    NUM_TO_ELEM,
    atomic_names,
    chemical_symbols,
)
from .lattice import Lattice

ElementLike = Union[str, int, "Element"]


# TODO: make this class an enumeration?
class Element:
    """
    Class representing an abtract chemical element, but no particular atom.
    This class gives access to elemental properties, like atomic number,
    atomic mass, full element name, etc.

    Parameters
    ----------
    element : str, int, or :class:`Element`
        Elemental symbol (e.g. "He"), element name (e.g. "Helium"),
        atomic number, or another `Element` instance.

    Raises
    ------
    ValueError : if the element is not valid.
    """

    valid_symbols = frozenset(chemical_symbols)
    valid_names = frozenset(atomic_names)

    def __init__(self, element: ElementLike, *args, **kwargs):
        if isinstance(element, int):
            try:
                element = NUM_TO_ELEM[element]
            except KeyError:
                raise ValueError(f"Atomic number {element} not supported.")
        elif isinstance(element, Element):
            element = element.symbol

        # At this point, `element` is a string
        element = str(element).title()

        valid_string_inputs = self.valid_names.union(self.valid_symbols)
        if element not in valid_string_inputs:
            raise ValueError(f"Element {element} is not valid.")

        if element in self.valid_symbols:
            self.element = element
        elif element in self.valid_names:
            self.element = NAME_TO_ELEM[element]

    def __str__(self) -> str:
        return self.symbol

    def __repr__(self) -> str:
        return f"< {self.name} >"

    def __eq__(self, other) -> bool:
        if isinstance(other, Element):
            return self.element == other.element
        return NotImplemented

    def __hash__(self) -> int:
        # Technically, if atomic_number is an int, hash(atomic_number) = atomic_number
        # However, just in case this won't be true in the future, we still use the hash() function
        return hash(self.atomic_number)

    @property
    def element_full(self) -> str:
        """Full element name, e.g. "Hydrogen" """
        return self.name

    @property
    def name(self) -> str:
        """Full element name, e.g. "Hydrogen" """
        return ELEM_TO_NAME[self.element]

    @property
    def symbol(self) -> str:
        """Elemental symbol, e.g. "He" """
        return self.element

    @property
    def atomic_number(self) -> int:
        """Atomic number"""
        return ELEM_TO_NUM[self.element]

    @property
    def mass(self) -> float:
        """Atomic mass [u]"""
        return ELEM_TO_MASS[self.element]

    @property
    def magnetic_moment_ground(self) -> float:
        """Ground state magnetic moment."""
        return ELEM_TO_MAGMOM.get(self.element, None)


class Atom(Element):
    """
    Container object for atomic data.

    Parameters
    ----------
    element : str or int
        Chemical element symbol or atomic number.
    coords : array-like, shape (3,)
        Coordinates of the atom in fractional form.
    lattice : Lattice or array-like, shape (3,3)
        Lattice on which the atom is positioned.
    displacement : array-like or None, optional
        Atomic maximum displacement [Angs].
    magmom : float, optional
        Magnetic moment. If None (default), the ground-state magnetic moment is used.
    occupancy : float, optional
        Fractional occupancy. If None (default), occupancy is set to 1.0.
    tag : int or None, optional
        Tag an atom with a unique identifier. Useful to keep track of atom order, for example
        in PWSCF output files. This is mostly for internal use.
    electronic_structure : ElectronicStructure or None, optional
        Electronic orbital structure for this atom. If `None` (default), the ground
        state for this element will be used.
    """

    # Because of the possibility of a large number of atoms (> 1e6), we use the __slots__
    # mechanism to make Atom objects as small as possible.
    __slots__ = (
        "element",
        "coords_fractional",
        "displacement",
        "magmom",
        "occupancy",
        "lattice",
        "electronic_structure",
    )

    def __init__(
        self,
        element: ElementLike,
        coords: ArrayLike,
        lattice: Optional[Union[Lattice, ArrayLike]] = None,
        displacement: ArrayLike = None,
        magmom: Optional[float] = None,
        occupancy: float = 1.0,
        tag: Optional[Any] = None,
        electronic_structure: Optional["ElectronicStructure"] = None,
        **kwargs,
    ):
        super().__init__(element=element)

        self.coords_fractional = np.asarray(coords, dtype=float)
        self.lattice = lattice or Lattice(np.eye(3))
        self.displacement = np.asarray(
            displacement if displacement is not None else (0, 0, 0), dtype=float
        )
        self.magmom = magmom or self.magnetic_moment_ground
        self.occupancy = occupancy
        self.tag = tag

        # We distinguish between default (None) and other electronic structures
        # So that the atoms can be represented appropriately in __repr__
        self.electronic_structure = (
            ElectronicStructure.ground_state(self.element) or electronic_structure
        )

    def __repr__(self) -> str:
        x, y, z = tuple(self.coords_fractional)
        r = f"< Atom {self.element:<2} @ ({x:.2f}, {y:.2f}, {z:.2f})"
        # No point in polluting the representation if the electronic structure was not
        # set (i.e. default)
        if self.electronic_structure != ElectronicStructure.ground_state(self.element):
            r += f" | [{str(self.electronic_structure)}]"
        r += " >"
        return r

    def __eq__(self, other) -> bool:
        if isinstance(other, Atom):
            return (
                super().__eq__(other)
                and (self.magmom == other.magmom)
                and distance_fractional(self, other) < 1e-3
                and (self.lattice == other.lattice)
                and np.allclose(self.displacement, other.displacement, atol=1e-3)
                and self.tag == other.tag
                and (self.electronic_structure == other.electronic_structure)
            )
        return False

    def __hash__(self) -> int:
        return hash(
            (
                super().__hash__(),
                self.magmom,
                tuple(np.round(self.coords_fractional, 4)),
                self.lattice,
                tuple(np.round(self.displacement, 3)),
                self.tag,
                hash(self.electronic_structure),
            )
        )

    def __lt__(self, other) -> bool:
        # Here's a cool fact:
        # Sorting n-tuples works as follows:
        #   1. Sort by the first elements
        #   2. If the first elements are equal, sort by the second elements;
        #   3. and so on
        # This means that by sorting atoms by (a.element, c1, c2 ,c3),
        # we can have a stable order (i.e. sorted by fractional coordinates)
        # but with elements grouped together.
        # The part where sorting is stable is important because of doctests,
        # which are quite literal.
        return (self.element, *self.coords_fractional) < (
            other.element,
            *other.coords_fractional,
        )

    @classmethod
    def from_ase(cls, atom) -> "Atom":
        """
        Returns an Atom instance from an ASE atom

        Parameters
        ----------
        atom : ase.Atom
        """
        lattice = np.eye(3)
        if atom.atoms is not None:
            lattice = np.array(atom.atoms.cell)

        return cls(
            element=atom.symbol,
            coords=frac_coords(atom.position, lattice),
            magmom=atom.magmom,
        )

    @property
    def coords_cartesian(self) -> np.ndarray:
        """
        Real-space position of the atom on the lattice, in Angstroms.

        Returns
        -------
        pos : `~numpy.ndarray`, shape (3,)
            Atomic position

        Raises
        ------
        RuntimeError : if this atom is not place on a lattice
        """
        return real_coords(self.coords_fractional, self.lattice.lattice_vectors)

    def transform(self, *matrices: ArrayLike) -> "Atom":
        """
        Return an Atom with fractional coordinates transformed according to symmetry operators.

        Parameters
        ----------
        matrices : ndarrays, shape {(3,3), (4,4)}
            Transformation matrices.

        Returns
        -------
        atm : Atom
            Transformed atom. The original atom is left unchanged.
        """
        coords_fractional = np.array(self.coords_fractional, copy=True)

        for matrix in matrices:
            coords_fractional = transform(matrix, coords_fractional)

        new_atom = deepcopy(self)
        new_atom.coords_fractional[:] = coords_fractional

        return new_atom

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """Returns an array [Z, x, y, z]"""
        # Not all keyword arguments can be passed to 'np.empty'
        kwargs.pop("copy", None)
        arr = np.empty(shape=(4,), *args, **kwargs)
        arr[0] = self.atomic_number
        arr[1::] = self.coords_fractional
        return arr


def real_coords(frac_coords: ArrayLike, lattice_vectors: ArrayLike) -> np.ndarray:
    """
    Calculates the real-space coordinates of the atom from fractional coordinates and lattice vectors.

    Parameters
    ----------
    frac_coords : array-like, shape (3,)
        Fractional coordinates
    lattice_vectors : list of ndarrays, shape (3,)
        Lattice vectors of the crystal.

    Returns
    -------
    coords : ndarray, shape (3,)
    """
    COB = change_of_basis(np.array(lattice_vectors), np.eye(3))
    return transform(COB, frac_coords)


def frac_coords(real_coords: ArrayLike, lattice_vectors: ArrayLike) -> np.ndarray:
    """
    Calculates the fractional coordinates of the atom from real-space coordinates and lattice vectors.

    Parameters
    ----------
    real_coords : array-like, shape (3,)
        Real-space coordinates
    lattice_vectors : list of ndarrays, shape (3,)
        Lattice vectors of the crystal.

    Returns
    -------
    coords : ndarray, shape (3,)
    """
    COB = change_of_basis(np.eye(3), np.array(lattice_vectors))
    return transform(COB, real_coords)


def distance_fractional(atm1: Atom, atm2: Atom) -> float:
    """
    Calculate the distance between two atoms in fractional coordinates.

    Parameters
    ----------
    atm1, atm2 : :class:`Atom`

    Returns
    -------
    dist : float
        Fractional distance between atoms.

    Raises
    ------
    RuntimeError : if atoms are not associated with the same lattice.
    """
    if atm1.lattice != atm2.lattice:
        raise RuntimeError(
            "Distance is undefined if atoms are sitting on different lattices."
        )

    return np.linalg.norm(atm1.coords_fractional - atm2.coords_fractional)


def distance_cartesian(atm1: Atom, atm2: Atom) -> float:
    """
    Calculate the distance between two atoms in cartesian coordinates.

    Parameters
    ----------
    atm1, atm2 : :class:`Atom`

    Returns
    -------
    dist : float
        Cartesian distance between atoms in Angstroms..

    Raises
    ------
    RuntimeError : if atoms are not associated with the same lattice.
    """
    if atm1.lattice != atm2.lattice:
        raise RuntimeError(
            "Distance is undefined if atoms are sitting on different lattices."
        )

    return np.linalg.norm(atm1.coords_cartesian - atm2.coords_cartesian)


def is_element(element: ElementLike) -> Callable[[Atom], bool]:
    """
    Create a function that checks whether an atom is of a certain element.

    Parameters
    ----------
    element : str, int, or Element
        Elemental symbol (e.g. "He"), atomic number, or Element instance.

    Returns
    -------
    func : callable
        Returns a function that can be used to check whether a `crystals.Atom`
        instance is of a certain element.

    Examples
    --------
    >>> is_vanadium = is_element('V') # is_vanadium is a function
    >>> atm = Atom('V', [0,0,0])
    >>> is_vanadium(atm)
    True
    """
    if not isinstance(element, Element):
        element = Element(element)

    def _is_element(atm):
        return atm.atomic_number == element.atomic_number

    return _is_element


@unique
class Orbital(Enum):
    """
    Enumeration of electronic orbitals, used to described atomic
    orbital structure.

    We note that `Orbital` instances are ordered according to the Madelung rule.

    Examples
    --------
    >>> Orbital("1s")
    <Orbital.one_s: '1s'>
    >>> Orbital.four_p == Orbital("4p")
    True

    """

    # It is important that the Orbitals are listed in the order that they
    # are filled (Madelung rule) because this is how __lt__ is defined.
    one_s = "1s"
    two_s = "2s"
    two_p = "2p"
    three_s = "3s"
    three_p = "3p"
    four_s = "4s"
    three_d = "3d"
    four_p = "4p"
    five_s = "5s"
    four_d = "4d"
    five_p = "5p"
    six_s = "6s"
    four_f = "4f"
    five_d = "5d"
    six_p = "6p"
    seven_s = "7s"
    five_f = "5f"
    six_d = "6d"
    seven_p = "7p"

    def __lt__(self, other: "Orbital") -> bool:
        madelung_rule = [v for k, v in Orbital.__members__.items()]
        return madelung_rule.index(self) < madelung_rule.index(other)

    @classmethod
    def maximum_electrons(cls, shell: Union[str, "Orbital"]) -> int:
        """
        Maximum number of electrons that can be placed in an orbital.

        Parameters
        ----------
        shell : Orbital or str

        Returns
        -------
        max : int
        """
        shell = Orbital(shell)
        maxima = {
            "s": 2,
            "p": 6,
            "d": 10,
            "f": 14,
        }
        return maxima[shell.value[-1]]


# To print superscript in electronic structures.
# Note that this requires UTF-8 output.
superscript_map = {
    "0": "⁰",
    "1": "¹",
    "2": "²",
    "3": "³",
    "4": "⁴",
    "5": "⁵",
    "6": "⁶",
    "7": "⁷",
    "8": "⁸",
    "9": "⁹",
}
superscript_trans = str.maketrans(
    "".join(superscript_map.keys()), "".join(superscript_map.values())
)


class ElectronicStructure:
    """
    Description of the atomic orbital structure.

    Parameters
    ----------
    shells : dict[Orbital,int]
        Dictionary containing the number of electrons in each Orbital, e.g. `{"1s": 2}`.

    Raises
    ------
    ValueError : if the electronic structure is not representable

    Examples
    --------
    Electronic structures can be specified by hand:

    >>> ElectronicStructure({"1s":2, "2s":2, "2p":2})
    < ElectronicStructure: 1s²2s²2p² >

    A shortcut exists for atomic ground states:

    >>> ElectronicStructure.ground_state("Ti")
    < ElectronicStructure: 1s²2s²2p⁶3s²3p⁶4s²3d² >

    Notes
    -----
    Shells are allowed to be filled our of order deliberately, given that unusual
    electronic structures can arise from ultrafast photoexcitation.
    """

    def __init__(self, shells: Dict[Union[str, Orbital], int]):
        shells = {Orbital(k): v for k, v in shells.items()}

        # Subclassing OrderedDict causes problems with pickling
        # Instead, we dress this class on top of an OrderedDict property.
        self._structure = OrderedDict([])
        for k, v in shells.items():
            self.__setitem__(k, v)

    def __iter__(self):
        yield from self._structure

    def __setitem__(self, key: Union[str, Orbital], value: int):
        # We check that the number of electrons in each Orbital does not
        # go above maximum possible.
        shell = Orbital(key)
        maximum_allowed_electrons = Orbital.maximum_electrons(shell)
        if value > maximum_allowed_electrons:
            raise ValueError(
                f"There cannot be {value} electrons in orbital {shell.value}"
            )
        self._structure.__setitem__(shell, value)

        # We ensure that orbital order is maintained
        # by re-creating the OrderedDict, filled
        self._structure = OrderedDict(sorted(self._structure.items()))

    def __getitem__(self, key: Union[str, Orbital]) -> int:
        # In case the key doesn't exist, we return 0
        # (i.e. 0 electrons in this orbital) because this allows
        # to add electrons in-place, e.g.:
        # >>> struct = ElectronicStructure({"1s":2})
        # >>> struct["2p"] += 1
        # even though there were no electrons there.
        key = Orbital(key)
        try:
            return self._structure.__getitem__(key)
        except KeyError:
            return 0

    def __str__(self) -> str:
        result = ""
        for shell, occ in self._structure.items():
            result += shell.value + f"{occ}".translate(superscript_trans)
        return result

    def __repr__(self) -> str:
        return f"< ElectronicStructure: {str(self)} >"

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: "ElectronicStructure") -> bool:
        return str(self) == str(other)

    @classmethod
    def ground_state(cls, element: ElementLike) -> "ElectronicStructure":
        """
        Ground state electronic structure for a particular element.

        Parameters
        ----------
        element : Element, str, or int
            Element, Symbol, or atomic number.

        Returns
        -------
        structure : ElectronicStructure
            Return the ground state electronic structure for a particular element.

        Examples
        --------
        >>> ElectronicStructure.ground_state("Ne")
        < ElectronicStructure: 1s²2s²2p⁶ >
        """
        element = Element(element)
        num_elec = element.atomic_number

        structure = dict()
        for shell in sorted(Orbital):
            shell_elec = min([Orbital.maximum_electrons(shell), num_elec])
            structure[shell] = shell_elec
            num_elec -= shell_elec

            if num_elec == 0:
                break

        return cls(structure)

    @property
    def outer_shell(self) -> Orbital:
        """The outermost shell, or valence orbital."""
        shells = list(self._structure.keys())
        return shells[-1]



# -*- coding: utf-8 -*-
import collections.abc as abc
import pickle
import pytest
from copy import deepcopy
from itertools import islice

import numpy as np

from crystals import Atom, AtomicStructure, Crystal

np.random.seed(23)


@pytest.fixture
def structure():
    substructure = AtomicStructure(atoms=[Atom("U", [0, 0, 0])])
    return AtomicStructure(
        atoms=[Atom("Ag", [0, 0, 0]), Atom("Ag", [1, 1, 1])],
        substructures=[substructure],
    )


def test_iteration(structure):
    """Test iteration of AtomicStructure yields from orphan atoms and substructure atoms alike"""
    elements = [atm.element for atm in structure]
    assert len(elements), 3


def test_addition_trivial(structure):
    """Test that the addition of two AtomicStructures, one being empty, works as expected"""
    addition = structure + AtomicStructure()
    assert addition == structure
    assert addition is not structure


def test_addition_uniqueness(structure):
    """Test that the addition of two AtomicStructures, works as expected regarding unique atoms"""
    assert structure + structure == structure


def test_addition(structure):
    """Test the addition of two different AtomicStructures works as expected."""
    new_struct = AtomicStructure(
        atoms=[Atom("U", [0, 1, 0])],
        substructures=[
            AtomicStructure(atoms=[Atom("Ag", [0.5, 0, 0]), Atom("Ag", [1, 0.3, 1])])
        ],
    )

    addition = structure + new_struct

    assert len(new_struct) + len(structure) == len(addition)
    assert len(new_struct.atoms) + len(structure.atoms) == len(addition.atoms)
    assert len(new_struct.substructures) + len(structure.substructures) == len(
        addition.substructures
    )


def test_addition_subclasses(structure):
    """Test that the addition of two subclass of AtomicStructures is preserved under addition."""

    class NewAtomicStructure(AtomicStructure):
        pass

    addition = NewAtomicStructure() + NewAtomicStructure()
    assert type(addition) is NewAtomicStructure


def test_truthiness(structure):
    """Test that empty AtomicStructures are falsey, and truthy otherwise."""
    empty_structure = AtomicStructure()
    assert not empty_structure

    assert structure


def test_trivial_transformation(structure):
    """Test that the identity transformation of an AtomicStructure works as expected."""
    transformed = structure.transform(np.eye(3))

    # transformed structure should be different, but equal, to original structure
    assert transformed is not structure
    assert transformed == structure


def test_transformations_inversions(structure):
    """Test that symmetry operations work as expected when inverted."""
    operator = np.random.random(size=(3, 3))
    inv_op = np.linalg.inv(operator)

    transformed1 = structure.transform(operator)
    transformed2 = transformed1.transform(inv_op)

    # transformed2 structure should be different, but equal, to original structure
    assert transformed2 is not structure
    assert transformed2 == structure


def test_transform_subclass(structure):
    """Test that the object returned by the transform() method is the
    same class as the method caller."""

    class NewAtomicStructure(AtomicStructure):
        pass

    structure = NewAtomicStructure(atoms=[Atom("Ag", [0, 0, 0]), Atom("Ag", [1, 1, 1])])
    transformed = structure.transform(np.eye(3))

    assert type(transformed) is type(structure)


def test_transformations_correctness(structure):
    """Test that AtomicStructure.transform() works as expected."""
    operator = 2 * np.eye(3)
    transformed = structure.transform(operator)

    expected_atoms = [atm.transform(operator) for atm in structure]

    for atm in expected_atoms:
        assert atm in transformed


def test_itersorted(structure):
    """Test that AtomicStructure.itersorted() works as expected"""
    sorted_from_structure = list(structure.itersorted())
    sorted_from_list = list(sorted(structure))

    assert sorted_from_structure == sorted_from_list


def test_chemical_composition_trivial(structure):
    """Test that AtomicStructure.chemical_composition works as expected"""
    expected = {"U": 1 / 3, "Ag": 2 / 3}
    assert structure.chemical_composition == expected


@pytest.mark.parametrize("name", islice(Crystal.builtins, 10))
def test_chemical_composition_add_to_unity(name):
    """Test that AtomicStructure.chemical_composition always adds up to 1"""
    # Faster to create a large atomic structure from a Crystal object
    # Testing for 10 crystal structures only
    structure = AtomicStructure(atoms=Crystal.from_database(name))
    assert round(abs(sum(structure.chemical_composition.values()) - 1), 7) == 0


def test_chemical_formula(structure):
    """Test that AtomicStructure.chemical_formula is working as expected."""
    assert structure.chemical_formula == "Ag2 U"


def test_chemical_formula_hill_notation(structure):
    """Test that the Hill notation, where elements are alphabetically ordered except C and H, which are first."""
    structure = AtomicStructure(
        atoms=[
            Atom("Ag", [0, 1, 0]),
            Atom("C", [0, 0, 0]),
            Atom("H", [0, 1, 0]),
            Atom("U", [1, 1, 1]),
        ]
    )
    assert structure.chemical_formula == "C H Ag U"


def test_length(structure):
    """Test the __len__ methods"""
    assert len(structure), 3


def test_containership_substructures(structure):
    """Test that containership works on substructure and atoms separately"""
    substructure = next(iter(structure.substructures))
    assert substructure in structure
    assert structure not in substructure


def test_containership_atoms(structure):
    """Test that atom containership testing is working, even in substructures"""
    atm = next(iter(structure))
    assert atm in structure


def test_equality(structure):
    """Test that AtomicStructure is equal to itself but not others"""
    substructure = next(iter(structure.substructures))
    assert structure == structure
    assert structure == deepcopy(structure)
    assert structure != substructure

    # Special case: make structures from Crystals
    c1 = Crystal.from_database("vo2-m1")
    c2 = deepcopy(c1)
    s1 = AtomicStructure(atoms=c1)
    s2 = AtomicStructure(atoms=c2.atoms)
    assert s1 == s2


def test_array(structure):
    """Test AtomicStructure.__array__"""
    arr = np.array(structure)
    assert arr.shape == (len(structure), 4)


def test_picklable(structure):
    """Test that Crystal instances can be pickled, and that the unpickled instance
    is identical to the source"""
    pickled = pickle.dumps(structure)
    unpickled = pickle.loads(pickled)
    assert structure == unpickled


def test_abstract_base_classes(structure):
    """Test that AtomicStructure fits with collections.abc module"""
    for abstract_base_class in (abc.Hashable, abc.Iterable, abc.Sized):
        assert isinstance(structure, abstract_base_class)


def test_satisfying(structure):
    """Test the AtomicStructure.satisfying method"""
    uranium = structure.satisfying(lambda a: a.element == "U")
    silver = structure.satisfying(lambda a: a.element == "Ag")

    assert len(uranium) == 1
    assert len(silver) == 2



# -*- coding: utf-8 -*-
"""
crystals command-line utilities
"""

import argparse
import sys
from pathlib import Path

from . import Crystal, __version__

constructors = {
    "database": Crystal.from_database,
    "cif": Crystal.from_cif,
    "pwscf": Crystal.from_pwscf,
    "cod": Crystal.from_cod,
    "pdb": Crystal.from_pdb,
    "mp": Crystal.from_mp,
    "poscar": Crystal.from_poscar,
}

INFO_HELP = "Display the crystallographic information related to a crystal file or database entry."

INFO_EXAMPLES = """
Example of a local CIF file:
> crystals info vo2.cif 
> crystals info vo2.cif --type cif

Example of a local PWSCF file:
> crystals info graphite.pwscf
> crystals info graphite.pwscf --type pwscf

Example of internal database entry:
> crystals info diamond
> crystals info diamond --type database

Example of Crystallography Open Database (COD) entry:
> crystals info 5000215 # Bismuth
> crystals info 5000215 --type cod

Example of Protein DataBank entry:
> crystals info 1fbb # Bacteriorhodopsin
> crystals info 1fbb --type pdb

Example of Materials Project entry (requires API key):
> crystals info Fe2O3 
> crystals info Fe2O3 --type mp

Example of a local POSCAR file:
> crystals info MgSiO3.direct.POSCAR
> crystals info MgSiO3.direct.POSCAR --type poscar
"""

INPUT_HELP = """Path to a file, or database ID. CIF files (*.cif) and PWSCF (*.pwscf) files are supported. 
Database IDs from the internal database, Crystallography Open Database (COD) and the Protein 
DataBank (PDB) are supported. """

parser = argparse.ArgumentParser(
    prog="crystals", description=f"crystals {__version__} command-line utilities"
)

subparsers = parser.add_subparsers(title="command", dest="command")

info_parser = subparsers.add_parser(
    "info",
    help=INFO_HELP,
    description=INFO_HELP,
    epilog=INFO_EXAMPLES,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
info_parser.add_argument("input", nargs=1, help=INPUT_HELP)

info_parser.add_argument(
    "--type",
    help="Type of input. By default, `crystals` will try to guess.",
    type=str,
    choices=[f"{itype}" for itype in constructors],
    required=False,
)


def guess_constructor(i):
    """
    Try to guess the appropriate Crystal constructor

    Returns
    -------
    cryst : Crystal

    Raises
    ------
    RuntimeError: if the input type could not be determined.
    """
    # Some easy heuristics first
    if str(i).endswith(".cif"):
        return Crystal.from_cif(i)
    elif str(i).endswith(".pwscf"):
        return Crystal.from_pwscf(i)
    elif str(i).startswith("mp-"):
        return Crystal.from_mp(i)
    elif "poscar" in str(i).lower():
        return Crystal.from_poscar(i)

    # We iterate over constructors that don't require network access first
    cryst = None
    for constructor in (
        Crystal.from_cif,
        Crystal.from_pwscf,
        Crystal.from_poscar,
        Crystal.from_database,
        Crystal.from_cod,
        Crystal.from_pdb,
        Crystal.from_mp,
    ):
        try:
            cryst = constructor(i)
        except:
            continue
        else:
            break
    if cryst is None:
        raise RuntimeError

    return cryst


def main():
    args = parser.parse_args()

    if args.command == "info":
        if args.type is not None:
            cryst = constructors[args.type](args.input[0])
        else:
            try:
                cryst = guess_constructor(args.input[0])
            except RuntimeError:
                print(
                    "Input type could not be determined. Try to use the --type option."
                )
                sys.exit(1)

        unitcell = (
            repr(cryst).replace("< ", "").replace(" >", "")
        )  # This is from python's object representation
        sym = f"""Symmetry information:
    International symbol 
                (short) ..... {cryst.international_symbol}
                (full) ...... {cryst.international_full}
    International number .... {cryst.international_number}
    Hermann-Mauguin symbol .. {cryst.hm_symbol}
    Pointgroup .............. {cryst.pointgroup}
    Hall Number ............. {cryst.hall_number}
    Centering ............... {cryst.centering}
    """

        print(unitcell)
        print(sym)
        sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Crystal structure indexing with the DirAx algorithm.
"""
from itertools import product
import numpy as np
from ..lattice import Lattice
from .common import IndexingError, row_echelon_form


def index_dirax(reflections, initial=None, length_bounds=(2, 20)):
    """
    Find the lattice associated with a list of
    reflections using the DirAx algorithm.

    .. versionadded:: 1.3.0

    Parameters
    ----------
    reflections : iterable of 3-tuple or ndarray, shape (N, 3)
        Iterable of reflections with their three-dimensional reciprocal space
        coordinates, or ndarray where each row is a reflections. Coordinates are
        in inverse Angstroms.
    initial : :class:`Lattice` or :class:`Crystal`, optional
        Initial guess for a lattice. The DirAx algorithm does not need an initial guess, but
        it can certainly help when many reflections are missing.
    length_bounds : 2-tuple of floats, optional
        Minimum and maximum lattice vector lengths to consider, in Angstrom.

    Returns
    -------
    indexed : :class:`Lattice`
        Lattice that best indexes `reflections`.
    indices : ndarray, shape (N, 3)
        Miller indices associated with each vector in `reflections`, in order.

    Raises
    ------
    IndexingError
        The indexing has failed.

    Notes
    -----
    This indexing routine is based on the reference below. The algorithm is well-suited to
    situations where reflections might be missing, or situations where the list of reflections
    contains "alien" reflections not associated with the lattice.

    Examples
    --------
    We generate reflections from a crystal structure and re-index it for demonstration
    purposes.

    >>> from crystals import index_dirax, Crystal
    >>> import numpy as np
    >>> graphite = Crystal.from_database('C')
    >>> # The list of reflections `qs` might be experimental measurements from either
    >>> # x-ray or electron diffraction.
    >>> qs = [graphite.scattering_vector(r) for r in graphite.bounded_reflections(bound=3.5)]
    >>> lattice, hkls = index_dirax(qs)
    >>> lattice # doctest: +SKIP
    < Lattice object with parameters 2.464Å, 2.464Å, 6.711Å, 90.00°, 90.00°, 120.00° >

    References
    ----------
    A. J. M. Duisenberg, Indexing in Single-Crystal Diffractometry with an Obstinate List of Reflections (1992),
    J. Appl. Cryst vol. 25 pp. 92 - 96
    """
    reflections = np.asarray(reflections, dtype=float)

    length_min, length_max = sorted(length_bounds)
    d_max = 2 * np.pi / length_min
    d_min = 2 * np.pi / length_max

    ns = np.arange(start=1, stop=13, step=1).reshape((-1, 1))

    potential_direct_vectors = set()

    # If a "guess" lattice is provided, we include its
    # lattice vectors as being high-priority.
    if initial is not None:
        for v in initial.lattice_vectors:
            potential_direct_vectors.add(LatVec(nf=len(reflections), vector=v))

    points = [np.squeeze(a) for a in np.vsplit(reflections, reflections.shape[0])]
    for a1, a2, a3 in product(points, repeat=3):
        normal = np.cross(a2 - a1, a3 - a1)
        if np.allclose(normal, 0, atol=1e-2):
            continue
        normal /= np.linalg.norm(normal)

        proj = np.sort(np.sum(reflections * normal[None, :], axis=1))
        d_star_start = np.diff(proj).max()
        if d_star_start <= 0:
            continue

        frac = proj / d_star_start / ns
        residues = np.sum(np.abs(frac - np.rint(frac)), axis=1)
        best_n = np.argmin(residues) + 1

        d_star = d_star_start / best_n
        if (d_star < d_min) or (d_star > d_max):
            continue

        frac_dist = proj / d_star
        nf = np.sum(np.isclose(frac_dist, np.rint(frac_dist), atol=1 / 24))

        t = 2 * np.pi * normal / d_star
        potential_direct_vectors.add(LatVec(nf, t))

    if len(potential_direct_vectors) == 0:
        raise IndexingError("No candidate lattice vectors could be determined.")

    max_nf = max((v.nf for v in potential_direct_vectors))
    acceptance_level = 0.8 * max_nf
    potential_direct_vectors = list(
        filter(lambda v: v.nf >= acceptance_level, potential_direct_vectors)
    )

    lattice = _find_basis(
        (v.vector for v in potential_direct_vectors), reflections=reflections
    )
    return lattice, lattice.miller_indices(reflections)


class LatVec:
    """
    Potential lattice vector, including the number indexed reflections `nf`
    """

    def __init__(self, nf, vector):
        self.nf = nf
        self.vector = np.around(vector, decimals=3)
        if np.sum(vector) < 0:
            self.vector *= -1

    def __hash__(self):
        return hash((self.nf, *tuple(self.vector)))

    def __repr__(self):
        return f"< LatVec: nf={self.nf}, vector={self.vector} >"


def _find_basis(vectors, reflections):
    """Find the shorted three linearly-independent vectors from a list."""
    vectors = sorted(vectors, key=np.linalg.norm)
    a1 = vectors.pop(0)
    try:
        index, a2 = next(
            (i, v)
            for i, v in enumerate(vectors)
            if np.linalg.norm(np.cross(a1, v)) >= 1
        )
    except StopIteration:
        raise IndexingError(
            "No set of three linearly-independent lattice vectors were found in the candidates."
        )
    vectors.pop(index)

    # The best way to find the last vector is to iterate through the remaining
    # vectors, from shortest to longest, until a matrix with the candidate
    # basis as rows has rank 3
    m = np.empty(shape=(3, 3), dtype=float)
    m[0, :] = a1
    m[1, :] = a2
    for v in vectors:
        m[2, :] = v
        if np.linalg.matrix_rank(m, tol=0.1) == 3:
            return Lattice(row_echelon_form([a1, a2, v]))

    raise IndexingError(
        "No set of three linearly-independent lattice vectors were found in the candidates."
    )



# -*- coding: utf-8 -*-
from copy import deepcopy
from random import random, seed
import pickle

import numpy as np
from itertools import islice

from crystals import (
    Atom,
    Orbital,
    ElectronicStructure,
    Element,
    Lattice,
    distance_cartesian,
    distance_fractional,
    is_element,
)
from crystals.atom_data import chemical_symbols
from crystals.affine import rotation_matrix
import pytest

seed(23)
np.random.seed(23)

# Madelung rule
MADELUNG = [
    "1s",
    "2s",
    "2p",
    "3s",
    "3p",
    "4s",
    "3d",
    "4p",
    "5s",
    "4d",
    "5p",
    "6s",
    "4f",
    "5d",
    "6p",
    "7s",
    "5f",
    "6d",
    "7p",
]


def random_transform():
    """Create a rotation matrix, around a random axis, of a random amount"""
    return rotation_matrix(random(), axis=np.random.random((3,)))


@pytest.mark.parametrize("symbol", chemical_symbols)
def test_element_build(symbol):
    """Test that all valid chemical symbols can be used to create an Element instance"""
    Element(symbol)


def test_element_invalid_element():
    """Test that an invalid chemical symbol will raise an error"""
    with pytest.raises(ValueError):
        Element("montreal")


@pytest.mark.parametrize("symbol", chemical_symbols)
def test_element_init_with_atomic_number(symbol):
    """Test that constructing an Element from a symbol or atomic number results in the same element"""
    from_symbol = Element(symbol)
    from_number = Element(from_symbol.atomic_number)
    assert from_number == from_symbol


def test_element_atomic_number_out_of_range():
    """Test that constructing an Element from an atomic number out-of-range raises a ValueError"""
    with pytest.raises(ValueError):
        Element(256)


def test_element_input_equivalence():
    """Test that the constructor for `Element` supports atomic numbers,
    strings, and other `Element`s"""
    expected = Element("H")

    assert Element("H") == expected
    assert Element("Hydrogen") == expected
    assert Element("hydrogen") == expected  # not case sensitive
    assert Element(1) == expected
    assert Element(expected) == expected


def test_atom_init():
    """Test that Atom can be instantiated with an element str or atomic number"""
    by_element = Atom("C", coords=(0, 0, 0))
    by_number = Atom(6, coords=(0, 0, 0))
    assert by_element == by_number


@pytest.mark.parametrize("tag", [None, 1, 2])
@pytest.mark.parametrize("symbol", islice(chemical_symbols, 50))
def test_atom_equality(symbol, tag):
    """Test __eq__ for atoms"""
    atom = Atom(symbol, coords=[0, 0, 0], tag=tag)
    other = deepcopy(atom)
    assert atom == atom
    assert atom == other
    assert hash(atom) == hash(other)

    other.coords_fractional = atom.coords_fractional + 1
    assert atom != other
    assert hash(atom) != hash(other)

    other = deepcopy(atom)
    other.tag = 5
    assert atom != other
    assert hash(atom) != hash(other)


@pytest.mark.parametrize(
    "atom", map(lambda s: Atom(s, [0, 0, 0]), islice(chemical_symbols, 50))
)
def test_atom_trivial_transform(atom):
    """Test Atom.transform() with the identity"""
    transformed = atom.transform(np.eye(3))

    assert transformed is not atom
    assert transformed == atom


def test_atom_subclasses_transform_preserved():
    """Test transform() preserves subclasses"""

    class NewAtom(Atom):
        pass

    atm = NewAtom("He", [0, 0, 0])
    transformed = atm.transform(np.eye(3))

    assert type(atm) is type(transformed)


@pytest.mark.parametrize(
    "atom", map(lambda s: Atom(s, [0, 0, 0]), islice(chemical_symbols, 50))
)
def test_atom_transform_back_and_forth(atom):
    """Test Atom.transform() with a random transformation back and forth"""

    transf = random_transform()
    transformed1 = atom.transform(transf)
    transformed2 = transformed1.transform(np.linalg.inv(transf))

    assert transformed2 is not atom
    assert transformed2 == atom


@pytest.mark.parametrize("atom", map(lambda s: Atom(s, [0, 0, 0]), chemical_symbols))
def test_atom_atom_array(atom):
    """Test that numpy.array(Atom(...)) works as expected"""
    arr = np.array(atom)
    assert arr.shape == (4,)
    assert arr[0] == atom.atomic_number
    assert np.allclose(arr[1::], atom.coords_fractional)


def test_atomic_distance_fractional():
    """Test the fractional distance between atoms"""
    atm1 = Atom("He", [0, 0, 0])
    atm2 = Atom("He", [1, 0, 0])
    assert distance_fractional(atm1, atm2) == 1
    assert distance_fractional(atm1, atm2) == distance_fractional(atm2, atm1)


def test_atomic_distance_cartesian():
    """Test the cartesian distance between atom"""
    lattice = Lattice(4 * np.eye(3))  # Cubic lattice side length 4 angs

    atm1 = Atom("He", [0, 0, 0], lattice=lattice)
    atm2 = Atom("He", [1, 0, 0], lattice=lattice)

    assert distance_cartesian(atm1, atm2) == 4.0


def test_atomic_distance_different_lattice():
    """Test that fractional and cartesian distances
    between atoms in different lattices raises an error."""
    lattice1 = Lattice(np.eye(3))
    lattice2 = Lattice(2 * np.eye(3))

    atm1 = Atom("He", [0, 0, 0], lattice=lattice1)
    atm2 = Atom("He", [1, 0, 0], lattice=lattice2)

    with pytest.raises(RuntimeError):
        distance_fractional(atm1, atm2)

    with pytest.raises(RuntimeError):
        distance_cartesian(atm1, atm2)


def test_is_element_input_types():
    """Test that is_element() works as expected for all
    supported input types"""
    atm = Atom("V", [0, 0, 0])

    assert is_element(atm.element)(atm)
    assert is_element(atm.atomic_number)(atm)
    assert is_element(atm)(atm)


def test_orbital_madelung_rule():
    """Test that the orbitals are listed in the Madelung rule order,
    which is the filling order."""

    enumeration = [shell.value for shell in Orbital]
    assert MADELUNG == enumeration


def test_orbital_maximum_electrons():
    """That that the maximum number of electrons per Orbital is as expected."""
    maxima = {"s": 2, "p": 6, "d": 10, "f": 14}
    for shell in Orbital:
        assert maxima[shell.value[-1]] == Orbital.maximum_electrons(shell)


@pytest.mark.parametrize("element, structure", (["C", {"1s": 2, "2s": 2, "2p": 2}],))
def test_electronic_structure_ground_state_examples(element, structure):
    gs = ElectronicStructure.ground_state(element)
    assert gs == ElectronicStructure(shells=structure)


@pytest.mark.parametrize("element", range(1, 24))
def test_electronic_structure_ground_state_num_electrons(element):
    gs = ElectronicStructure.ground_state(element)
    num_electrons = sum(gs[k] for k in gs)
    assert num_electrons == element


def test_electronic_structure_maximum_electrons():
    """Test that an error is raised for impossible electronic structures."""
    with pytest.raises(ValueError):
        ElectronicStructure({"1s": 3})


def test_electronic_structure_pickable():
    """Test that ElectronicStructure instances are pickable."""
    structure = ElectronicStructure.ground_state("W")
    assert structure == pickle.loads(pickle.dumps(structure))


def test_electronic_structure_missing_orbital():
    """Test that 'missing' orbitals will return 0 electrons"""
    struct = ElectronicStructure.ground_state("He")
    assert struct["2p"] == 0


def test_electronic_structure_orbital_keys():
    """Test that orbital occupancies can be accessed either with strings or Orbital"""
    struct = ElectronicStructure.ground_state("He")
    assert struct["1s"] == struct[Orbital.one_s]


def test_electronic_structure_modification_in_place():
    """Test that ElectronicStructure instances can be modified in-place"""
    struct = ElectronicStructure({"1s": 2, "2s": 2, "2p": 4})
    expected = ElectronicStructure({"1s": 2, "2s": 2, "2p": 3, "4s": 1})
    struct["2p"] -= 1
    struct["4s"] += 1
    assert struct == expected


def test_electronic_structure_valence_shell():
    """Test that the outermost shell is as expected"""
    struct = ElectronicStructure({"1s": 2})
    assert struct.outer_shell == Orbital("1s")

    # Intentionally omitting 2s
    struct = ElectronicStructure({"1s": 2, "2p": 4})
    assert struct.outer_shell == Orbital("2p")



# -*- coding: utf-8 -*-
from typing import Iterable

from numpy.typing import ArrayLike

from .affine import affine_map
from .base import AtomicStructure


class Residue(AtomicStructure):
    """
    Fundamental building block of a protein: amino acid residue.

    Parameters
    ----------
    atoms : iterable, optional
        Atoms not attached to a substructure.
    name : str
        Residue name, e.g. "GLY"
    sequence_number : int
        Sequence number within a protein.
    """

    def __init__(
        self, atoms: AtomicStructure, name: str, sequence_number: int, **kwargs
    ):
        super().__init__(atoms=atoms, **kwargs)
        self.name = name
        self.sequence_number = int(sequence_number)

    def transform(self, *operators: ArrayLike) -> "Residue":
        """
        Return a transformed Residue based on symmetry operators.

        operators : iterable of array_like
            Symmetry operators, either 3x3 or 4x4.

        Returns
        -------
        transformed : Residue
            New structure.
        """
        operators = tuple(map(affine_map, operators))

        transformed_atoms = set()
        for atom in self.atoms:
            for operator in operators:
                transformed_atoms.add(atom.transform(operator))

        # We defer construction to the current class. Therefore, subclasses of AtomicStructure
        # will transform into their own class
        return self.__class__(
            atoms=transformed_atoms,
            name=self.name,
            sequence_number=self.sequence_number,
        )


# Helper class to not repeat code betwee Helix and Sheet
class SecondaryStructure(AtomicStructure):
    """
    Abstract container representing protein secondary structures (i.e. helices, sheets, etc.)

    Parameters
    ----------
    residues : iterable, optional
        Residues making this structure.
    sequence_number : int
        Sequence number within a protein.
    """

    def __init__(self, residues: Iterable[Residue], sequence_number: int, **kwargs):
        super().__init__(substructures=residues, **kwargs)
        self.sequence_number = int(sequence_number)

    @property
    def residues(self) -> Iterable[Residue]:
        """Residues making this structure."""
        return self.substructures

    def transform(self, *operators: ArrayLike) -> "SecondaryStructure":
        """
        Return a transformed SecondaryStructure based on symmetry operators.

        operators : iterable of array_like
            Symmetry operators, either 3x3 or 4x4.

        Returns
        -------
        transformed : SecondaryStructure
            New structure.
        """
        operators = tuple(map(affine_map, operators))

        # Substructures are recursively transformed
        transformed_substructures = set()
        for substructure in self.substructures:
            transformed_substructures.add(substructure.transform(*operators))

        # We defer construction to the current class. Therefore, subclasses of AtomicStructure
        # will transform into their own class
        return self.__class__(
            residues=transformed_substructures, sequence_number=self.sequence_number
        )


class Helix(SecondaryStructure):
    """
    Container representing a protein helix.

    Parameters
    ----------
    residues : iterable, optional
        Residues making this helix.
    sequence_number : int
        Sequence number within a protein.
    """

    def transform(self, *operators: ArrayLike) -> "Helix":
        """
        Return a transformed helix based on symmetry operators.

        operators : iterable of array_like
            Symmetry operators, either 3x3 or 4x4.

        Returns
        -------
        transformed : Helix
            New structure.
        """
        return super().transform(*operators)


class Sheet(SecondaryStructure):
    """
    Container representing a protein sheet.

    Parameters
    ----------
    residues : iterable, optional
        Residues making this sheet.
    sequence_number : int
        Sequence number within a protein.
    """

    def transform(self, *operators: ArrayLike) -> "Sheet":
        """
        Return a transformed sheet based on symmetry operators.

        operators : iterable of array_like
            Symmetry operators, either 3x3 or 4x4.

        Returns
        -------
        transformed : Sheet
            New structure.
        """
        return super().transform(*operators)



# -*- coding: utf-8 -*-

import tempfile
from contextlib import suppress
from itertools import islice
from pathlib import Path
import numpy as np
from tempfile import TemporaryDirectory

import pytest
from crystals import Crystal, Atom, __version__
from crystals.writers import CIF_HEADER

try:
    import ase
except ImportError:
    WITH_ASE = False
else:
    WITH_ASE = True


@pytest.mark.skipif(not WITH_ASE, reason="ASE not installed or importable")
@pytest.mark.parametrize("name", islice(Crystal.builtins, 10))
def test_ase_atoms_construction(name):
    """Test that ase_atoms returns without error"""
    crystal = Crystal.from_database(name)
    to_ase = crystal.to_ase()
    assert len(crystal) == len(to_ase)


@pytest.mark.skipif(not WITH_ASE, reason="ASE not installed or importable")
@pytest.mark.parametrize("name", islice(Crystal.builtins, 10))
def test_ase_atoms_back_and_forth(name):
    """Test conversion to and from ase Atoms"""
    crystal = Crystal.from_database(name)
    to_ase = crystal.to_ase()
    crystal2 = Crystal.from_ase(to_ase)

    # ase has different handling of coordinates which can lead to
    # rounding beyond 1e-3. Therefore, we cannot compare directly sets
    # assertSetEqual(set(crystal), set(crystal2))
    assert len(crystal) == len(crystal2)


@pytest.mark.parametrize("name", Crystal.builtins)
def test_cif_writer_idempotence(name):
    """Test that conversion to CIF of a structure loaded from CIF is idempotent."""
    # Note that this test can be flaky due to symmetry determination in spglib 2+
    with suppress(RuntimeError):
        cryst = Crystal.from_database(name)
        with tempfile.TemporaryDirectory() as temp_dir:
            f = Path(temp_dir) / "temp.cif"
            cryst.to_cif(f)
            cryst2 = Crystal.from_cif(f)
            assert cryst == cryst2


def test_cif_writer_writes_version():
    """Test that the `crystals` version is contained in the CIF header."""
    # See https://github.com/LaurentRDC/crystals/issues/14
    assert f"`crystals` {__version__}" in CIF_HEADER


def test_supercell_preserved_in_cif():
    """See issue #13."""
    a = np.asarray([3, 0.2, 0], dtype=float)
    b = np.asarray([0.4, 8, 0], dtype=float)
    c = np.asarray([0, 0, 9], dtype=float)

    dimer = Crystal(
        unitcell=[Atom("C", coords=(0.5, 0, 0)), Atom("C", coords=(0, 0, 0))],
        lattice_vectors=[a * 2, b, c],
    )
    with TemporaryDirectory() as tmpdir:
        dimer.to_cif(Path(tmpdir) / "dimer.cif")
        from_file = Crystal.from_cif(Path(tmpdir) / "dimer.cif")

    assert set(tuple(atm.coords_fractional) for atm in dimer.unitcell) == set(
        tuple(atm.coords_fractional) for atm in from_file.unitcell
    )


@pytest.mark.parametrize("name", Crystal.builtins)
def test_vasp_writer_idempotence(name):
    """Test that conversion to VASP of a structure loaded from VASP is idempotent."""
    # Testing on all built-in structure assures us that corner cases
    # are taken care of.
    cryst = Crystal.from_database(name)
    with tempfile.TemporaryDirectory() as temp_dir:
        f = Path(temp_dir) / "temp.POSCAR"
        cryst.to_poscar(f)
        cryst2 = Crystal.from_poscar(f)
        assert cryst == cryst2



# -*- coding: utf-8 -*-
"""
Basic definitions common to all indexers.
"""
import numpy as np


class IndexingError(RuntimeError):
    """
    Indexing has failed.

    .. versionadded :: 1.3.0
    """

    pass


def ratio_indexed(lattice, reflections):
    """
    Determine the proportion of `reflections` that are indexed by `lattice`, i.e.
    that have integer Miller indices.

    Parameters
    ----------
    lattice : :class:`Lattice` or :class:`Crystal`

    reflections : iterable of 3-tuple or ndarray, shape (N, 3)
        Iterable of reflections with their three-dimensional reciprocal space
        coordinates, or ndarray where each row is a reflections. Coordinates are
        in inverse Angstroms.

    Returns
    -------
    ratio : float
        Ratio of reflections that are indexed correctly, in the [0, 1] range.
    """
    reflections = np.asarray(reflections, dtype=float)

    hkl = lattice.miller_indices(reflections)
    return (
        np.sum(np.all(np.isclose(hkl - np.rint(hkl), 0, atol=0.1), axis=1))
        / reflections.shape[0]
    )


def row_echelon_form(A):
    """
    Transform the matrix `A` in row echelon form.

    .. versionadded :: 1.3.0
    """
    # Algorithm adapted from:
    #   https://johnfoster.pge.utexas.edu/numerical-methods-book/LinearAlgebra_DirectSolvers.html
    A = np.array(A, copy=True)

    for i in range(A.shape[0]):
        pivot = i + np.abs(A[i:, i]).argmax()

        # Swapping rows to make the maximal entry the
        # pivot (if needed).
        if pivot != i:
            A[[pivot, i]] = A[[i, pivot]]

        # Eliminate all entries below the pivot
        factor = A[i + 1 :, i] / A[i, i]
        A[i + 1 :] -= factor[:, None] * A[i]

    return A



# -*- coding: utf-8 -*-
from collections import namedtuple
from copy import deepcopy
from enum import Enum, unique
from functools import lru_cache
from itertools import chain, combinations, groupby, islice, product, starmap
from operator import attrgetter
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, Optional, Set, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike
from spglib import (
    get_error_message,
    get_spacegroup_type,
    get_symmetry,
    get_symmetry_dataset,
    standardize_cell,
)

from .affine import affine_map, change_of_basis
from .atom import Atom
from .atom_data import chemical_symbols
from .base import AtomicStructure
from .lattice import Lattice
from .parsers import (
    CIFParser,
    CODParser,
    MPJParser,
    PDBParser,
    POSCARParser,
    PWSCFParser,
)
from .spg_data import Hall2HM
from .writers import ase_atoms, write_cif, write_poscar, write_xyz

CIF_ENTRIES = frozenset((Path(__file__).parent / "cifs").glob("*.cif"))
LONGEST_CHEMICAL_SYMBOL = max(len(s) for s in chemical_symbols)

ACCEPTABLE_SITE_SYMMETRIES = frozenset(
    {
        "wyckoffs",
        "crystallographic_orbits",
        "equivalent_atoms",
    }
)

is_atom = lambda a: isinstance(a, Atom)
is_structure = lambda s: isinstance(s, AtomicStructure)


class Crystal(AtomicStructure, Lattice):
    """
    The :class:`Crystal` class is a set-like container that represent
    crystalline structures. In addition to constructing the :class:`Crystal`
    object yourself, other constructors are also available
    (and preferred):

    * :meth:`Crystal.from_cif`: create an instance from a CIF file;

    * :meth:`Crystal.from_pdb`: create an instance from a Protein Data Bank entry;

    * :meth:`Crystal.from_database`: create an instance from the internal database of CIF files;

    * :meth:`Crystal.from_cod`: create an instance from a Crystallography Open Database entry;

    * :meth:`Crystal.from_mp`: create an instance from the Materials Project database;

    * :meth:`Crystal.from_pwscf`: create an instance from the output of the PWSCF program;

    * :meth:`Crystal.from_ase`: create an instance from an ``ase.Atoms`` instance;

    * :meth:`Crystal.from_poscar`: create an instance from VASP POSCAR files.

    Parameters
    ----------
    unitcell : iterable of :class:`Atom` or :class:`AtomicStructure`
        Unit cell atoms or substructures. It is assumed that the atoms are
        in fractional coordinates.
    lattice_vectors : iterable of array_like
        Lattice vectors. If ``lattice_vectors`` is provided as a 3x3 array, it
        is assumed that each lattice vector is a row.
    source : str or None, optional
        Provenance, e.g. filename. Only used for bookkeeping.
    """

    builtins = frozenset(map(attrgetter("stem"), CIF_ENTRIES))

    def __init__(
        self,
        unitcell: Iterable[Union[Atom, AtomicStructure]],
        lattice_vectors: ArrayLike,
        source: Optional[str] = None,
        **kwargs,
    ):
        unitcell = list(unitcell)
        # Atoms need to be modified BEFORE they are fed to the constructor
        # of AtomicStructure
        for atom in filter(is_atom, unitcell):
            atom.lattice = Lattice(lattice_vectors)

        super().__init__(
            atoms=filter(is_atom, unitcell),
            substructures=filter(is_structure, unitcell),
            lattice_vectors=lattice_vectors,
            **kwargs,
        )

        self.source = source

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Crystal):
            # The explicit comparison of lattices sidesteps the problem
            # of evaluating NotImplemented in a bool context, which has been deprecated
            # as of Python 3.9
            return (
                Lattice(self.lattice_vectors) == Lattice(other.lattice_vectors)
            ) and super(AtomicStructure, self).__eq__(other)
        return NotImplemented

    def __hash__(self):
        return super(AtomicStructure, self).__hash__() | super(Lattice, self).__hash__()

    @property
    def unitcell(self) -> Iterator[Atom]:
        """Generator of atoms forming the crystal unit cell."""
        return super().__iter__()

    # This operation is very heavy; better to cache it than recalculate.
    @lru_cache(maxsize=1)
    def asymmetric_cell(self) -> Set[Atom]:
        """
        Calculates the asymmetric cell that generates the crystal unit cell.

        .. versionadded:: 1.2.0
        """
        return symmetry_reduction(self.unitcell, self.symmetry_operations())

    @classmethod
    @lru_cache(maxsize=len(builtins))
    def from_cif(cls, path: PathLike, **kwargs):
        """
        Returns a Crystal object created from a CIF 1.0, 1.1 or 2.0 file.
        Keyword arguments are passed to the Crystal constructor.

        Parameters
        ----------
        path : path-like
            File path
        """
        with CIFParser(filename=path) as parser:
            return cls(
                unitcell=symmetry_expansion(
                    parser.atoms(), parser.symmetry_operators()
                ),
                lattice_vectors=parser.lattice_vectors(),
                source=str(path),
                **kwargs,
            )

    @classmethod
    def from_database(cls, name: str, **kwargs):
        """
        Returns a Crystal object create from the internal CIF database.
        Keyword arguments are passed to the class constructor.

        Parameters
        ----------
        name : path-like
            Name of the database entry. Available items can be retrieved from `Crystal.builtins`
        """
        if name not in cls.builtins:
            raise ValueError(
                f"Entry {name} is not available in the database. See `Crystal.builtins` for valid entries."
            )

        path = Path(__file__).parent / "cifs" / (name + ".cif")
        return cls.from_cif(path, **kwargs)

    @classmethod
    def from_cod(
        cls,
        num: int,
        revision: Optional[int] = None,
        download_dir: Optional[PathLike] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Returns a Crystal object built from the Crystallography Open Database.
        Keyword arguments are passed to the class constructor.

        Parameters
        ----------
        num : int
            COD identification number.
        revision : int or None, optional
            Revision number. If None (default), the latest revision is used.
        download_dir : path-like object, optional
            Directory where to save the CIF file. Default is a local folder in the current directory
        overwrite : bool, optional
            Whether or not to overwrite files in cache if they exist. If no revision
            number is provided, files will always be overwritten.
        """
        with CODParser(num, revision, download_dir, overwrite) as parser:
            return cls(
                unitcell=symmetry_expansion(
                    parser.atoms(), parser.symmetry_operators()
                ),
                lattice_vectors=parser.lattice_vectors(),
                source=f"COD num:{num} rev:{revision}",
                **kwargs,
            )

    @classmethod
    def from_mp(
        cls,
        query: str,
        api_key: Optional[str] = None,
        download_dir: Optional[PathLike] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Returns a Crystal object built from the Materials Project.
        Keyword arguments are passed to the class constructor.

        Parameters
        ----------
        query : str
            The query can be a Materials Project material id (e.g., `"mp-1234"`), a
            formula, e.g. (`"Fe2O3"`), or a chemical system ("-" separated list of elemments,
            e.g., `"Li-Fe-O"`).
        api_key : str or None, optional
            An API key for accessing the Materials Project REST interface.
            Please obtain your API key at https://www.materialsproject.org/dashboard.
            If `None` (default), ``crystals`` will look for your API key in the
            `MATERIALS_PROJECT_API_KEY` environment variable.
        download_dir : path-like object or None, optional
            Directory where to save the CIF file. This is used for caching.
        overwrite : bool, optional
            Whether or not to overwrite files in cache if they exist. If True,
            a new file will be downloaded, possibly overwriting previously-downloaded file.
        """
        with MPJParser(
            query=query,
            api_key=api_key,
            download_dir=download_dir,
            overwrite=overwrite,
            **kwargs,
        ) as parser:
            return cls(
                unitcell=symmetry_expansion(
                    parser.atoms(), parser.symmetry_operators()
                ),
                lattice_vectors=parser.lattice_vectors(),
                source=f"Materials Project ID: {parser.material_id}",
                **kwargs,
            )

    @classmethod
    def from_pdb(
        cls,
        ID: str,
        download_dir: Optional[PathLike] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Returns a Crystal object created from a Protein DataBank entry.
        Keyword arguments are passed to the class constructor.

        Parameters
        ----------
        ID : str
            Protein DataBank identification. The correct .pdb file will be downloaded,
            cached and parsed.
        download_dir : path-like object, optional
            Directory where to save the PDB file.
        overwrite : bool, optional
            Whether or not to overwrite files in cache if they exist. If no revision
            number is provided, files will always be overwritten.
        """
        with PDBParser(ID=ID, download_dir=download_dir, overwrite=overwrite) as parser:
            return cls(
                unitcell=symmetry_expansion(
                    parser.residues(), parser.symmetry_operators()
                ),
                lattice_vectors=parser.lattice_vectors(),
                source=parser.filename,
                **kwargs,
            )

    @classmethod
    def from_pwscf(cls, path: PathLike, **kwargs):
        """
        Returns a Crystal object created from an output file of PWSCF.
        Keyword arguments are passed to the class constructor.

        Parameters
        ----------
        path : path-like
            File path
        """
        with PWSCFParser(path) as parser:
            return cls(
                unitcell=parser.atoms(),
                lattice_vectors=parser.lattice_vectors(),
                source=parser.filename,
                **kwargs,
            )

    @classmethod
    def from_ase(cls, atoms, **kwargs):
        """
        Returns a Crystal object created from an ASE Atoms object.
        Keyword arguments are passed to the class constructor.

        Parameters
        ----------
        atoms : ase.Atoms
            Atoms group.
        """
        lattice_vectors = atoms.get_cell()

        return cls(
            unitcell=[Atom.from_ase(atm) for atm in atoms],
            lattice_vectors=lattice_vectors,
            **kwargs,
        )

    @classmethod
    def from_poscar(cls, path: PathLike, **kwargs):
        """
        Returns a Crystal object created from a VASP's POSCAR file.
        Keyword arguments are passed to the class constructor.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        path : path-like
            File path
        """
        with POSCARParser(path) as parser:
            return cls(
                unitcell=parser.atoms(),
                lattice_vectors=parser.lattice_vectors(),
                source=parser.filename,
                **kwargs,
            )

    def _spglib_cell(self) -> np.ndarray:
        """Returns an array in spglib's cell format."""
        # To get symmetry information, we only give spglib the unit cell atoms
        # This way, all spglib-related methods (like symmetry()) will act on the unit cell only.
        # This distinction is important for Crystal subclasses, like Supercell.
        unitcell = np.stack([np.asarray(atm) for atm in self.unitcell])
        return np.array(self.lattice_vectors), unitcell[:, 1:], unitcell[:, 0]

    @classmethod
    def _from_spglib_cell(
        cls,
        lattice_vectors: np.ndarray,
        scaled_positions: np.ndarray,
        numbers: np.ndarray,
        **kwargs,
    ):
        """
        Build a Crystal object from the return value of many SPGLIB routines.

        Parameters
        ----------
        lattice_vectors : ndarray, shape (3,3)
            Lattice vectors
        scaled_positions : ndarray, shape (N, 3)
            Fractional atomic positions, row-wise.
        numbers : ndarray, shape (N,)
            Atomic numbers, associated with each row of `scaled_positions`
        """
        atoms = [
            Atom(int(Z), coords=coords) for Z, coords in zip(numbers, scaled_positions)
        ]
        # Preserve whatever subclass this object already is
        # This is important because some properties can be extracted from
        # source files (e.g. PWSCF output files)
        return cls(unitcell=atoms, lattice_vectors=lattice_vectors, **kwargs)

    def primitive(self, symprec: float = 1e-2) -> "Crystal":
        """
        Returns a Crystal object in the primitive unit cell.

        Parameters
        ----------
        symprec : float, optional
            Symmetry-search distance tolerance in Cartesian coordinates [Angstroms].

        Returns
        -------
        primitive : Crystal
            Crystal with primitive cell. Even if the crystal already has a primitive
            cell, a new crystal is returned.

        Raises
        ------
        RuntimeError : If primitive cell could not be found.

        Notes
        -----
        Optional atomic properties (e.g magnetic moment) might be lost in the reduction.
        """
        search = standardize_cell(
            self._spglib_cell(), to_primitive=True, no_idealize=True, symprec=symprec
        )
        if search is None:
            raise RuntimeError("Primitive cell could not be found.")

        return self._from_spglib_cell(*search, source=self.source)

    def ideal(self, symprec: float = 1e-2) -> "Crystal":
        """
        Returns a Crystal object with an idealized unit cell.

        Parameters
        ----------
        symprec : float, optional
            Symmetry-search distance tolerance in Cartesian coordinates [Angstroms].

        Returns
        -------
        ideal : Crystal
            Crystal with idealized cell.

        Raises
        ------
        RuntimeError : If an ideal cell could not be found.

        Notes
        -----
        Optional atomic properties (e.g magnetic moment) might be lost in the symmetrization.
        """
        search = standardize_cell(
            self._spglib_cell(), to_primitive=True, no_idealize=False, symprec=symprec
        )
        if search is None:
            raise RuntimeError("Ideal cell could not be found.")

        return self._from_spglib_cell(*search, source=self.source)

    def supercell(self, n1: int, n2: int, n3: int) -> "Supercell":
        """
        Create a supercell from this crystal, i.e. an atomic structure where the crystal unit cell
        is duplicated along lattice vectors.

        Parameters
        ----------
        n1, n2, n3 : int
            Repeat along the `a1`, `a2`, and `a3` lattice vectors. For example,
            ``(1, 1, 1)`` represents the trivial supercell.

        Returns
        -------
        cell : Supercell
            Iterable of `crystals.Atom` objects following the supercell dimensions.
        """
        multicell = list()
        for atm in self:
            for factors in product(range(n1), range(n2), range(n3)):
                fractional_offset = np.asarray(factors)
                newatom = deepcopy(atm)
                newatom.coords_fractional += fractional_offset
                multicell.append(newatom)

        return Supercell(
            unitcell=multicell,
            lattice_vectors=self.lattice_vectors,
            dimensions=(n1, n2, n3),
        )

    def symmetry(
        self, symprec: float = 1e-2, angle_tolerance: float = -1.0
    ) -> Dict[str, Any]:
        """
        Returns a dictionary containing space-group information. This information
        is computed from the crystal unit cell.

        Parameters
        ----------
        symprec : float, optional
            Symmetry-search distance tolerance in Cartesian coordinates [Angstroms].
        angle_tolerance: float, optional
            Symmetry-search tolerance in degrees. If the value is negative (default),
            an internally optimized routine is used to judge symmetry.

        Returns
        -------
        info : dict
            Dictionary of space-group information. The following keys are available:

                * ``'international_symbol'``: International Tables of Crystallography
                  space-group symbol (short);

                * ``'international_full'``: International Tables of
                  Crystallography space-group full symbol;

                * ``'hall_symbol'`` : Hall symbol;

                * ``'hm_symbol'`` : Hermann-Mauguin symbol;

                * ``'centering'``: Centering-type ("P", "F", etc.);

                * ``'pointgroup'`` : International Tables of
                  Crystallography point-group;

                * ``'international_number'`` : International Tables of
                  Crystallography space-group number (between 1 and 230);

                * ``'hall_number'`` : Hall number (between 1 and 531).

        Raises
        ------
        RuntimeError : if symmetry-determination has not succeeded.

        Notes
        -----
        Note that crystals generated from the Protein Data Bank are often incomplete;
        in such cases the space-group information will be incorrect.
        """
        dataset = get_symmetry_dataset(
            cell=self._spglib_cell(), symprec=symprec, angle_tolerance=angle_tolerance
        )

        err_msg = get_error_message()
        if err_msg != "no error":
            raise RuntimeError(
                f"[SPGLIB] Symmetry-determination has returned the following error: {err_msg}"
            )

        if dataset is None:
            raise RuntimeError("[SPGLIB] Symmetry-determination has not found a match.")

        spg_type = get_spacegroup_type(dataset["hall_number"])
        try:
            hm_symbol = Hall2HM[dataset["hall"]]
        except KeyError:
            raise RuntimeError(
                f"[SPGLIB] Unknown Hermann-Mauguin symbol associated to Hall symbol \"{dataset['hall']}\""
            )

        # We do not distinguish between base-centered "A", "B", and "C"
        # "A" and "B" are translated to "C"
        centering = CenteringType(
            hm_symbol[0] if hm_symbol[0] not in {"A", "B"} else "C"
        )
        return {
            "international_symbol": dataset["international"],
            "hall_symbol": dataset["hall"],
            "hm_symbol": hm_symbol,
            "centering": centering,
            "international_number": dataset["number"],
            "hall_number": dataset["hall_number"],
            "international_full": spg_type["international_full"],
            "pointgroup": spg_type["pointgroup_international"],
        }

    def symmetry_operations(self, symprec: float = 1e-2) -> Iterable[np.ndarray]:
        """
        Get the symmetry operations that the crystal unit cell respects. These symmetry operations
        are expressed in fractional coordinates.

        Parameters
        ----------
        symprec : float, optional
            Symmetry-search distance tolerance in Cartesian coordinates [Angstroms].

        Returns
        -------
        sym_ops : iterable of array_like, shapes (4,4)
            Iterable of affine transforms, where ``m[:3,:3]`` is the rotation part,
            while ``m[:3,-1]`` is the translation.

        Raises
        ------
        RuntimeError : if symmetry-determination has not succeeded.

        See also
        --------
        Crystal.reciprocal_symmetry_operations : symmetry operations in reciprocal basis
        """
        dataset = get_symmetry(cell=self._spglib_cell(), symprec=symprec)

        def _to_affine(r, t):
            """Convert rotation and translation into single 4x4 affine transformation"""
            m = np.eye(4)
            m[:3, :3] = r
            m[:3, -1] = t
            return m

        return [
            _to_affine(r, t)
            for r, t in zip(dataset["rotations"], dataset["translations"])
        ]

    def reciprocal_symmetry_operations(
        self, symprec: float = 1e-2
    ) -> Iterable[np.ndarray]:
        """
        Get the symmetry operations that the reciprocal unit cell respects. These symmetry operations
        are expressed in reciprocal fractional coordinates.

        Parameters
        ----------
        symprec : float, optional
            Symmetry-search distance tolerance in Cartesian coordinates [Angstroms].

        Returns
        -------
        sym_ops : iterable of array_like, shapes (4,4)
            Iterable of affine transforms, where ``m[:3,:3]`` is the rotation part,
            while ``m[:3,-1]`` is the translation.

        Raises
        ------
        RuntimeError : if symmetry-determination has not succeeded.

        See also
        --------
        Crystal.symmetry_operations : symmetry operations in lattice basis
        """
        transformations = self.symmetry_operations(symprec=symprec)

        # Change of basis matrices allow to express
        # transformations in other bases
        to_reciprocal = change_of_basis(
            np.array(self.lattice_vectors), np.array(self.reciprocal_vectors)
        )
        from_reciprocal = np.linalg.inv(to_reciprocal)

        cast = lambda m: to_reciprocal @ m @ from_reciprocal

        # Pack and unpack rotation and translation from/to affine transform
        # TODO: this is ugly
        def pack(r, t):
            m = np.eye(4)
            m[:3, :3] = r
            m[:3, -1] = t
            return m

        ops = list()
        for m in transformations:
            r, t = m[:3, :3], m[:3, -1]
            ops.append(pack(cast(r), cast(t)))
        return ops

    @property
    def international_symbol(self) -> str:
        """International Tables of Crystallography space-group short symbol."""
        return self.symmetry()["international_symbol"]

    @property
    def international_full(self) -> str:
        """International Tables of Crystallography space-group full symbol."""
        return self.symmetry()["international_full"]

    @property
    def hall_symbol(self) -> str:
        """Hall symbol."""
        return self.symmetry()["hall_symbol"]

    @property
    def hm_symbol(self) -> str:
        """Hermann-Mauguin symbol."""
        return self.symmetry()["hm_symbol"]

    @property
    def pointgroup(self) -> str:
        """International Tables of Crystallography point-group."""
        return self.symmetry()["pointgroup"]

    @property
    def international_number(self) -> int:
        """International Tables of Crystallography space-group number (between 1 and 230)."""
        return self.symmetry()["international_number"]

    @property
    def hall_number(self) -> int:
        """Hall number (between 1 and 531)."""
        return self.symmetry()["hall_number"]

    @property
    def centering(self) -> "CenteringType":
        """Centering type of this crystals."""
        return self.symmetry()["centering"]

    # TODO: interpolate the docstring to specify the allowed values of `by`.
    #       f-strings cannot be used as docstrings unfortunately.
    def groupby(
        self,
        by: str,
        symprec: float = 1e-2,
        angle_tolerance: float = -1.0,
    ) -> Dict[Any, Iterable[Atom]]:
        """
        Group unit cell atoms by some measure of site symmetry, for example Wyckoff letters
        or crystallographic orbits.

        .. versionadded:: 1.6.0

        Parameters
        ----------
        by : str
            Measure of site symmetry by which to group atoms. Can be one of "wyckoffs", "crystallographic_orbits",
            or "equivalent_atoms". See the documentation of `spglib` for a description of these quantities.
        symprec : float, optional
            Symmetry-search distance tolerance in Cartesian coordinates [Angstroms].
        angle_tolerance: float, optional
            Symmetry-search tolerance in degrees. If the value is negative (default),
            an internally optimized routine is used to judge symmetry.

        Returns
        -------
        groups: dict
            Dictionary where each key is a distinct site symmetry (e.g. a distinct orbit), and values are
            iterable of :class:`Atom` which share the same site symmetry.

        Raises
        ------
        RuntimeError
            If symmetry-determination has not succeeded.
        ValueError
            If `by` is not one of the supported values.

        Notes
        -----
        Note that crystals generated from the Protein Data Bank are often incomplete;
        in such cases the site symmetry information will be incorrect.
        """
        if by not in ACCEPTABLE_SITE_SYMMETRIES:
            raise ValueError(
                f"Cannot group atoms by site-symmetry '{by}'. Expected one of: {ACCEPTABLE_SITE_SYMMETRIES}."
            )

        # Iterating over the atoms of a `Crystal` has an undefined order. As such,
        # we need to first 'freeze' the iteration order, and create a spglib-compatible
        # cell *with this order*. That's why we can't re-use `Crystal._spglib_cell`.
        sortkey = lambda atm: np.linalg.norm(atm.coords_fractional)
        atoms = sorted(self.unitcell, key=sortkey)

        unitcell_array = np.stack([np.asarray(atm) for atm in atoms])
        spglib_cell = (
            np.array(self.lattice_vectors),
            unitcell_array[:, 1:],
            unitcell_array[:, 0],
        )

        symmetry_dataset = get_symmetry_dataset(
            cell=spglib_cell, symprec=symprec, angle_tolerance=angle_tolerance
        )

        mksite = namedtuple("Site", ("atom", *ACCEPTABLE_SITE_SYMMETRIES))

        sites = starmap(
            mksite,
            zip(
                atoms,
                *(symmetry_dataset[measure] for measure in ACCEPTABLE_SITE_SYMMETRIES),
            ),
        )

        by = attrgetter(by)
        return {
            k: sorted((site.atom for site in v), key=sortkey)
            for k, v in groupby(iterable=sorted(sites, key=by), key=by)
        }

    def indexed_by(self, lattice: Union[Lattice, np.ndarray]) -> "Crystal":
        """
        Return a crystal structure, indexed by another lattice/crystal structure.

        Parameters
        ----------
        lattice : :class:`Crystal`, :class:`Lattice`, or array_like, shape (3,3)
            Lattice or Crystal by which to index the structure. ``lattice`` can also be a 3x3 array,
            where every row is a lattice vector.

        Returns
        -------
        crystal : :class:`Crystal`
            New crystal, indexed by ``lattice``.
        """
        # TODO: example
        if not isinstance(lattice, (Crystal, Lattice)):
            lattice = Lattice(lattice)

        old_basis = self.lattice_vectors
        new_basis = lattice.lattice_vectors

        cob = change_of_basis(old_basis, new_basis)
        return self.__class__(
            unitcell=(atm.transform(cob) for atm in self),
            lattice_vectors=new_basis,
            source=self.source,
        )

    def __str__(self) -> str:
        """String representation of this instance. Atoms may be omitted."""
        return self._to_string(natoms=10)

    def __repr__(self) -> str:
        """Verbose string representation of this instance."""
        return self._to_string(natoms=len(self))

    def _to_string(self, natoms: int) -> str:
        """Generate a string representation of this Crystal. Only include
        a maximum of `natoms` if provided."""

        # Note : Crystal subclasses need not override this method
        # since the class name is dynamically determined
        rep = f"< {self.__class__.__name__} object with following unit cell:"
        atoms = islice(sorted(self), natoms)

        # Note that repr(Atom(...)) includes these '< ... >'
        # We remove those for cleaner string representation
        rep += "".join(
            "\n    " + repr(atm).replace("<", "").replace(">", "").strip()
            for atm in atoms
        )

        num_omitted_atms = len(self) - natoms
        if num_omitted_atms > 0:
            rep += f"\n      ... omitting {num_omitted_atms:d} atoms ..."
            rep += "\n      ... use repr() to show the full cell ... "

        # Lattice parameters are split between lengths and angles
        a, b, c, alpha, beta, gamma = self.lattice_parameters
        rep += "\nLattice parameters:"
        rep += f"\n    a={a:.3f}Å, b={b:.3f}Å, c={c:.3f}Å"
        rep += f"\n    α={alpha:.3f}°, β={beta:.3f}°, γ={gamma:.3f}°"

        # Show stochiometric information
        # It is important that the chemical symbols are presented in order
        # so that doctests are reproducible
        rep += "\nChemical composition:"
        for chem_symbol, composition in sorted(
            self.chemical_composition.items(), key=lambda t: t[0]
        ):
            rep += f"\n    {chem_symbol.rjust(LONGEST_CHEMICAL_SYMBOL)}: {100 * composition:.3f}%"
        rep += " >"
        return rep

    def to_cif(self, filename: PathLike):
        """
        Convert this :class:`Crystal` instance to a CIF file.

        Note that some information may be lost in the translation. However, we guarantee that
        reading a structure from a file, and then writing back to the same format is idempotent.

        Parameters
        ----------
        filename : path-like
            Path to a file. If the file already exists, it will be overwritten.

        See also
        --------
        Crystal.to_xyz : write a structure to a `.xyz` file.
        Crystal.to_ase : convert a structure into an ``ase.Atoms`` object.
        """
        return write_cif(self, filename)

    def to_xyz(self, filename: PathLike):
        """
        Convert this :class:`Crystal` instance to a XYZ file.

        Note that some information may be lost in the translation. However, we guarantee that
        reading a structure from a file, and then writing back to the same format is idempotent.

        Parameters
        ----------
        filename : path-like
            Path to a file. If the file already exists, it will be overwritten.

        See also
        --------
        Crystal.to_cif : write a structure to a `.cif` file.
        Crystal.to_ase : convert a structure into an ``ase.Atoms`` object.
        """
        return write_xyz(self, filename)

    def to_ase(self, **kwargs):
        """
        Convert a into an :class:`ase.Atoms` object. Keyword arguments are passed
        to :class:`ase.Atoms` constructor.

        Note that some information may be lost in the translation. However, we guarantee that
        reading a structure from a file, and then writing back to the same format is idempotent.

        Returns
        -------
        atoms : ase.Atoms
            Group of atoms ready for ASE's routines.

        Raises
        ------
        ImportError : If ASE is not installed

        See also
        --------
        Crystal.to_cif : write a structure to a `.cif` file.
        Crystal.to_xyz : write a structure to a `.xyz` file.
        """
        return ase_atoms(self)

    def to_poscar(self, filename: PathLike, **kwargs):
        """
        Convert this :class:`Crystal` instance to a POSCAR file.
        Keyword arguments are passed to :meth:`writers.write_poscar`.

        Note that some information may be lost in the translation. However, we guarantee that
        reading a structure from a file, and then writing back to the same format is idempotent.

        Parameters
        ----------
        filename : path-like
            Path to a file. If the file already exists, it will be overwritten.
        kwargs:
            Keyword arguments are passed to :meth:`writers.write_poscar`.
        """
        return write_poscar(self, filename, **kwargs)


class Supercell(Crystal):
    """
    The :class:`Supercell` class is a set-like container that represents a
    supercell of crystalline structures. It has all the same attributes as a :class:`Crystal`, with
    the addition of :attr:`Supercell.scaled_lattice_vectors` and :attr:`Supercell.dimensions`.

    It is strongly recommended that you do not instantiate a :class:`Supercell` by hand, but rather
    create a :class:`Crystal` object and use the :meth:`Crystal.supercell` method.

    To iterate over all atoms in the supercell, use this object as an iterable.
    """

    def __init__(self, dimensions: Tuple[int, int, int], **kwargs):
        super().__init__(**kwargs)
        self.dimensions = dimensions

    @property
    def scaled_lattice_vectors(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """3-tuple of lattice vectors scaled by the dimensions of the supercell. If the supercell is (2x2x2) unit cells,
        then the scaled lattice vectors are all 2x of the unit cell lattice vectors.

        .. versionadded:: 1.5.0
        """
        n1, n2, n3 = self.dimensions
        return (n1 * self.a1, n2 * self.a2, n3 * self.a3)


@unique
class CenteringType(Enum):
    """
    Enumeration of possible centering types. Together with the lattice system,
    these centering types defined all 14 Bravais lattices in 3D.

    The possible centering types are:

    * ``'P'`` : Primitive

    * ``'I'`` : Body-centered

    * ``'F'`` : Face-centered

    * ``'C'`` : Base-centered

    * ``'R'`` : Rhombohedral in hexagonal setting.
    """

    primitive = "P"
    base_centered = "C"
    body_centered = "I"
    face_centered = "F"
    rhombohedral = "R"


def symmetry_expansion(
    atoms: Iterable[Union[Atom, AtomicStructure]],
    symmetry_operators: Iterable[ArrayLike],
) -> Iterator[Union[Atom, AtomicStructure]]:
    """
    Generate a set of unique atoms and structures from an asymmetric cell
    and symmetry operators.

    Parameters
    ----------
    atoms : iterable of :class:`Atom` and/or :class:`AtomicStructure`
        Asymmetric unit cell atoms. It is assumed that the atomic
        coordinates are in fractional form. Transformations work
        the same way for :class:`Atom` objects and :class:`AtomicStructure`
        objects: a copy is made and moved to the symmetric location.
    symmetry_operators : iterable of array_like
        Symmetry operators that generate the full unit cell.

    Yields
    ------
    it : :class:`Atom` and/or :class:`AtomicStructure`
        Appropriately-transformed object. Original objects are left untouched.

    See Also
    --------
    symmetry_reduction : Determine the asymmetric cell that can generate a unit cell.
    """
    # TODO: provide ability to reduce to primitive, niggli_reduce, etc.
    #       using spglib?

    is_identity = lambda op: np.allclose(op, np.eye(4))
    symmetry_operators = [
        m for m in map(affine_map, symmetry_operators) if not is_identity(m)
    ]

    unique_atoms = set([])
    for atm in filter(is_atom, atoms):
        # At least one atom is kept
        unique_atoms.add(atm)
        for sym_op in symmetry_operators:
            new = atm.transform(sym_op)
            new.coords_fractional[:] = np.mod(new.coords_fractional, 1)
            unique_atoms.add(new)

    def _normalize_fractional_coordinates(struct):
        """Ensure that all atoms in this structure have fractional coordinates that are valid, i.e.
        in [0, 1). Atoms are changed in-place."""
        for atm in struct.atoms:
            atm.coords_fractional[:] = np.mod(atm.coords_fractional, 1)

        for substruct in struct.substructures:
            _normalize_fractional_coordinates(substruct)

    unique_structures = set([])
    for structure in filter(is_structure, atoms):
        unique_structures.add(structure)
        for sym_op in symmetry_operators:
            new = structure.transform(sym_op)
            unique_structures.add(_normalize_fractional_coordinates(new))

    yield from unique_atoms
    yield from unique_structures


def symmetry_reduction(
    unitcell: Iterable[Union[Atom, AtomicStructure]],
    symmetry_operators: Iterable[ArrayLike],
) -> Iterator[Union[Atom, AtomicStructure]]:
    """
    Determine the asymmetric cell that generates `unitcell` when combined with
    symmetry operations. Effectively, this function is the reciprocal operation
    to `symmetry_expansion`.

    Note that the resulting asymmetric cell is not unique.

    .. versionadded:: 1.2.0

    Parameters
    ----------
    unitcell : iterable of :class:`Atom` and/or :class:`AtomicStructure`
        Unit cell. It is assumed that the atomic
        coordinates are in fractional form.
    symmetry_operators : iterable of array_like
        Symmetry operators that generate the full unit cell.

    Returns
    -------
    asym_cell: set of :class:`Atom` and/or :class:`AtomicStructure`
        Asymmetric cell.

    See Also
    --------
    symmetry_expansion : Expand the asymmetric cell into a unit cell.
    """
    # NOTE: This function does the dumb thing of trying to recreate the input
    #       by searching through all the possibilities.
    #       By nature, it is very slow; scaling like n^2 (where n is the unit cell length)
    #       However, other approaches that I tried were not reliable. For example,
    #       using inverse symmetry operators would result in numerical instabilities.
    #       If you are reading this and have a better idea, please submit a pull request :).

    unitcell = set(unitcell)

    # The powerset of the unit cell is the set of all possible sets
    # made from elements in `unitcell`
    powerset = chain.from_iterable(
        combinations(unitcell, r) for r in range(1, len(unitcell) + 1)
    )

    # We sort by length so that the resulting asymmetric cell is the smallest
    # possible one.
    for asym_cell in sorted(powerset, key=len):
        reconstituted = set(symmetry_expansion(asym_cell, symmetry_operators))
        if reconstituted == unitcell:
            return set(asym_cell)

    # Base case
    return unitcell



# -*- coding: utf-8 -*-
import tempfile
from copy import deepcopy
from functools import lru_cache
from itertools import islice, chain
from pathlib import Path
import socket
from contextlib import suppress

import numpy as np
from crystals import Atom, AtomicStructure, CenteringType, Crystal
from crystals.affine import translation_matrix
from crystals.crystal import (
    symmetry_expansion,
    symmetry_reduction,
    ACCEPTABLE_SITE_SYMMETRIES,
)
import pytest

np.random.seed(23)


@lru_cache(maxsize=1)
def connection_available():
    """Returns whether or not an internet connection is available"""
    with suppress(OSError), socket.create_connection(("www.google.com", 80)):
        return True
    return False


def test_symmetry_graphite():
    """Test that Crystal.symmetry() works correctly for graphite"""
    c = Crystal.from_database("C")
    info = c.symmetry(1e-1)

    supposed = {
        "international_number": 194,
        "hall_number": 488,
        "international_symbol": "P6_3/mmc",
        "international_full": "P 6_3/m 2/m 2/c",
        "hall_symbol": "-P 6c 2c",
        "hm_symbol": "P63/mmc",
        "centering": CenteringType("P"),
        "pointgroup": "6/mmm",
    }

    assert info == supposed


@pytest.mark.parametrize("name", Crystal.builtins)
def test_primitive_for_builtins(name):
    """Test that all built-in crystal have a primitive cell"""
    c = Crystal.from_database(name)
    prim = c.primitive(symprec=0.1)
    assert len(prim) <= len(c)


def test_primitive_preserves_subclass():
    """Check that Crystal returned by Crystal.primitive() preserve subclass"""

    class TestCrystal(Crystal):
        pass

    c = TestCrystal.from_database("C")
    prim = c.primitive()
    assert TestCrystal == type(prim)


@pytest.mark.parametrize("name", Crystal.builtins)
def test_ideal_for_builtins(name):
    """Test that all built-in crystal have an ideal cell"""
    # This will raise an error if no idealized cell is found
    c = Crystal.from_database(name).ideal()


def test_ideal_preserves_subclass():
    """Check that Crystal returned by Crystal.ideal() preserve subclass"""

    class TestCrystal(Crystal):
        pass

    c = TestCrystal.from_database("C")
    ideal = c.ideal()
    assert TestCrystal == type(ideal)


@pytest.mark.parametrize("name", Crystal.builtins)
def test_symmetry_operations(name):
    """Test that the symmetry operations output makes sense"""
    identity = np.eye(4)

    c = Crystal.from_database(name)
    symops = c.symmetry_operations()
    assert np.allclose(identity, symops[0])


@pytest.mark.parametrize("name", Crystal.builtins)
def test_reciprocal_symmetry_operations(name):
    """Test that the reciprocal symmetry operations output makes sense"""
    identity = np.eye(4)

    c = Crystal.from_database(name)
    symops = c.reciprocal_symmetry_operations()
    assert np.allclose(identity, symops[0])


@pytest.mark.parametrize("name", Crystal.builtins)
def test_str_vs_repr(name):
    """Test that str and repr are workign as expected"""
    c = Crystal.from_database(name)

    # If small crystal, repr and str should be the same
    if len(c) <= 10:
        assert repr(c) == str(c)
    else:
        assert repr(c) != str(c)


def test_equality():
    """Test that __eq__ works as expected"""
    c1 = Crystal.from_database("Pu-alpha")
    c2 = deepcopy(c1)
    assert c1 == c2

    c3 = Crystal.from_database("Pu-epsilon")
    assert c1 != c3


def test_containership():
    """Test that __contains__ works as expected"""
    graphite = Crystal.from_database("C")
    carbon = next(iter(graphite))  # Pick any atom from the crystal

    assert carbon in graphite


@pytest.mark.parametrize("name", Crystal.builtins)
def test_builtins(name):
    """Test that all names in Crystal.builtins build without errors,
    and that Crystal.source is correctly recorded."""
    c = Crystal.from_database(name)

    assert name in c.source


def test_builtins_wrong_name():
    """Test that a name not in Crystal.builtins will raise a ValueError"""
    with pytest.raises(ValueError):
        Crystal.from_database("___")


def test_substructure_preservation():
    """Test that initializing a crystal with substructures preserves the substructures"""
    atoms = [Atom("Ag", [0, 0, 0]), Atom("Ag", [1, 1, 1])]
    substructures = [AtomicStructure(atoms=[Atom("U", [0, 0, 0])])]
    c = Crystal(unitcell=atoms + substructures, lattice_vectors=np.eye(3))

    assert len(c) == 3
    assert substructures[0] in c.substructures


@pytest.mark.skipif(
    not connection_available(), reason="Internet connection is required."
)
def test_from_pdb():
    """Test Crystal.from_pdb constructor"""
    c = Crystal.from_pdb("1fbb")
    assert "1fbb" in c.source


@pytest.mark.skipif(
    not connection_available(), reason="Internet connection is required."
)
def test_from_cod():
    """Test building a Crystal object from the COD"""
    # revision = None and latest revision should give the same Crystal
    c = Crystal.from_cod(1521124)
    c2 = Crystal.from_cod(1521124, revision=176429)

    assert c == c2


@pytest.mark.skipif(
    not connection_available(), reason="Internet connection is required."
)
def test_from_cod_new_dir():
    """Test that a cache dir is created by Crystal.from_cod"""
    with tempfile.TemporaryDirectory() as temp_dir:
        download_dir = Path(temp_dir) / "test_cod"
        assert not download_dir.exists()
        c = Crystal.from_cod(1521124, download_dir=download_dir)
        assert download_dir.exists()


@pytest.mark.parametrize("name", islice(Crystal.builtins, 20))
def test_supercell_constructors(name):
    """Test Supercell constructors for varyous 'builtin' structures"""
    c = Crystal.from_database(name)
    s = c.supercell(2, 2, 2)

    assert len(s) == 8 * len(c)


@pytest.mark.parametrize("n1", [1, 2, 3])
@pytest.mark.parametrize("n2", [1, 2, 3])
@pytest.mark.parametrize("n3", [1, 2, 3])
def test_supercell_scaled_lattice_vectors(n1, n2, n3):
    """Test the `scaled_lattice_vectors` attribute of the `Supercell` class."""
    c = Crystal.from_database("C")
    a1, a2, a3 = c.lattice_vectors
    sa1, sa2, sa3 = c.supercell(n1, n2, n3).scaled_lattice_vectors

    assert np.allclose(n1 * a1, sa1)
    assert np.allclose(n2 * a2, sa2)
    assert np.allclose(n3 * a3, sa3)


@pytest.mark.parametrize("tag", [None, 1, 2])
@pytest.mark.parametrize("occupancy", [1.0, 1.5, 2])
@pytest.mark.parametrize("symbol", ["H", "He", "C"])
def test_supercell_preserve_attributes(tag, symbol, occupancy):
    """Test that the `Crystal.supercell` method preserves atom attributes (see issue #9)"""
    atom = Atom(element=symbol, coords=[0, 0, 0], tag=tag, occupancy=occupancy)
    c = Crystal(unitcell=[atom], lattice_vectors=3 * np.eye(3))
    s = c.supercell(2, 2, 2)

    for supercell_atom in s:
        assert supercell_atom.tag == atom.tag
        assert supercell_atom.symbol == atom.symbol


# The two tested site symmetry measures aren't *always* equivalent, but often are
@pytest.mark.parametrize(
    "site_symmetry_measure", ["crystallographic_orbits", "equivalent_atoms"]
)
@pytest.mark.parametrize(
    "crystal, num_distinct_sites", (["C", 2], ["vo2-m1", 3], ["vo2-rutile", 2])
)
def test_groupby(site_symmetry_measure, crystal, num_distinct_sites):
    """Test the Crystal.groupby method with known structures."""
    cryst = Crystal.from_database(crystal)
    gps = cryst.groupby(site_symmetry_measure)
    assert len(gps) == num_distinct_sites


@pytest.mark.parametrize("by", ACCEPTABLE_SITE_SYMMETRIES)
@pytest.mark.parametrize("crystal", Crystal.builtins)
def test_groupby_invariants(by, crystal):
    """Test that the grouping of atoms contains all atoms from the source crystal, and only those atoms."""
    cryst = Crystal.from_database(crystal)
    gps = cryst.groupby(by=by)
    assert set(chain.from_iterable(gps.values())) == set(cryst)


def test_groupby_unknown_symmetry():
    """Test that unknown symmetries passed to `Crystal.groupby` raise a ValueError."""
    cryst = Crystal.from_database("diamond")
    with pytest.raises(ValueError):
        cryst.groupby("this should not work")


def test_indexed_by_trivial_reindexing():
    """Test re-indexing a crystal by itself"""
    c1 = Crystal.from_database("Pu-gamma")
    c2 = c1.indexed_by(c1)

    assert c1 == c2


def test_symmetry_reduction_trivial():
    """Test that the symmetry_reduction function returns the unit cell when
    there is only one possibility."""
    ucell = set([Atom("H", coords=[0, 0, 0])])
    symops = [np.eye(4)]
    asym_cell = symmetry_reduction(ucell, symops)
    assert ucell == asym_cell


def test_symmetry_reduction_simple_translation():
    """Test that symmetry_reduction works on a unitcell where two atoms are
    linked by a translation"""
    symops = [np.eye(4), translation_matrix([0, 0, 1 / 3])]
    asym_cell = set([Atom("H", coords=[0, 0, 0])])
    unitcell = set(symmetry_expansion(asym_cell, symmetry_operators=symops))

    asym_cell2 = symmetry_reduction(unitcell, symops)
    assert asym_cell == asym_cell2


@pytest.mark.parametrize("name", ["vo2-m1", "Os", "Na"])
def test_symmetry_reduction_reciprocity_with_symmetry_expansion(name):
    """Test that symmetry_reduction is reciprocal to symmetry_expansion"""
    cryst = Crystal.from_database(name)
    asym_cell = symmetry_reduction(cryst.unitcell, cryst.symmetry_operations())
    ucell = set(symmetry_expansion(asym_cell, cryst.symmetry_operations()))
    assert set(cryst.unitcell) == ucell



# -*- coding: utf-8 -*-
import numpy as np

from crystals import Crystal, Lattice, IndexingError, index_dirax
import pytest

np.random.seed(2021)


# The structures to test have been chosen so that it doesn't take too long.
@pytest.mark.parametrize(
    "name,bound", zip(["Pu-epsilon", "C", "vo2-m1", "BaTiO3_cubic"], [2, 3, 2, 2])
)
def test_dirax_indexing_ideal(name, bound):
    """
    Test that indexing always succeeds with an ideal
    set of reflections: no noise, no alien reflections.
    """
    cryst = Crystal.from_database(name)
    refls = [cryst.scattering_vector(r) for r in cryst.bounded_reflections(bound=bound)]
    lat, hkls = index_dirax(refls)

    assert np.allclose(hkls - np.rint(hkls), 0, atol=0.001)


@pytest.mark.parametrize("name", ["Pu-epsilon", "C", "vo2-m1", "BaTiO3_cubic"])
def test_dirax_indexing_initial_guess(name):
    """
    Test that indexing succeeds with an initial guess and very few reflections.
    """
    cryst = Crystal.from_database(name)
    # We restrict the number of reflections to a single (0,0,0); with
    # the initial guess, indexing should still succeed!
    lat, _ = index_dirax([(0, 0, 0)], initial=cryst)

    assert np.allclose(lat.lattice_parameters, cryst.lattice_parameters, atol=1)


@pytest.mark.parametrize(
    "name,bound", zip(["Pu-epsilon", "C", "vo2-m1", "BaTiO3_cubic"], [2, 3, 2, 2])
)
def test_dirax_indexing_alien_reflections(name, bound):
    """
    Test that indexing always succeeds even with 20% of alien reflection.
    """
    cryst = Crystal.from_database(name)
    refls = [cryst.scattering_vector(r) for r in cryst.bounded_reflections(bound=bound)]
    num_aliens = len(refls) // 5
    aliens = [
        cryst.lattice_parameters[0] * np.random.random(size=(3,))
        for _ in range(num_aliens)
    ]
    lat, hkls = index_dirax(refls + aliens)
    # The alien reflections will not be indexed correctly, of course
    hkls = hkls[:num_aliens]
    assert np.allclose(hkls - np.rint(hkls), 0, atol=0.01)


@pytest.mark.parametrize(
    "name,bound", zip(["Pu-epsilon", "C", "vo2-m1", "BaTiO3_cubic"], [2, 3, 2, 2])
)
def test_dirax_indexing_noise(name, bound):
    """
    Test that indexing always succeeds despite noise in reflection positions.
    """
    np.random.seed(0)

    cryst = Crystal.from_database(name)
    refls = [
        cryst.scattering_vector(r) + np.random.normal(0, scale=0.01, size=(3,))
        for r in cryst.bounded_reflections(bound=bound)
    ]
    lat, hkls = index_dirax(refls)
    assert np.allclose(hkls - np.rint(hkls), 0, atol=0.1)


def test_dirax_indexing_length_bound():
    """
    Test that indexing fails as expected if the lattice length bounds are too restrictive.
    """
    cryst = Crystal.from_database("C")
    refls = [cryst.scattering_vector(r) for r in cryst.bounded_reflections(bound=2)]
    # We restrict the number of reflections to a single (0,0,0); with
    # the initial guess, indexing should still succeed!
    with pytest.raises(IndexingError):
        index_dirax(
            refls,
            length_bounds=(0.01, 2),
        )
